# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from abc import abstractmethod
from contextlib import AbstractContextManager
from typing import List


class AbstractBase(AbstractContextManager):
    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    def __exit__(self, *args, **kwargs):
        pass

    @abstractmethod
    def lattice(self, obj):
        pass

    @abstractmethod
    def atom_sites(self, obj):
        pass

    @abstractmethod
    def atom_site_aniso(self, obj):
        pass

    @abstractmethod
    def atom_site_sus(self, obj):
        pass

    @abstractmethod
    def symmetry(self, obj):
        pass

    @abstractmethod
    def structure(self, obj):
        pass

    @abstractmethod
    def structures(self, obj):
        pass


class AbstractStructureReader(AbstractBase):
    """
    Abstract class for reading structures.
    """

    @abstractmethod
    def read(self, input, *args, **kwargs) -> List[str]:
        """
        Reads a structure from a file.

        :param filename: The filename to read from.
        :return: The structure.
        """
        pass


class AbstractStructureWriter(AbstractBase):
    """
    Abstract class for writing structures.
    """

    @abstractmethod
    def write(self, structure, *args, **kwargs) -> None:
        """
        Writes a structure to a file.

        :param structure: The structure to write.
        :param filename: The filename to write to.
        """
        pass


class AbstractStructureParser:
    def __init__(self, reading_context_class, writing_context_class):
        self._reading_context_class = reading_context_class
        self._writing_context_class = writing_context_class

    def writer(self, *args, **kwargs):
        return self._writing_context_class(*args, **kwargs)

    def reader(self, *args, **kwargs):
        return self._reading_context_class(*args, **kwargs)

    def open(self, *args, rw: str = 'r', **kwargs):
        if rw == 'r':
            return self.reader(*args, **kwargs)
        elif rw == 'w':
            return self.writer(*args, **kwargs)
        else:
            raise ValueError('rw must be either "r" or "w"')

    def __call__(self, *args, rw: str = 'r', **kwargs):
        return self.open(*args, rw=rw, **kwargs)



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import NoReturn
from typing import Tuple

from easycrystallography.Components.SpaceGroup import SpaceGroup as _SpaceGroup
from easycrystallography.Symmetry.SymOp import SymmOp

from .template import CIF_Template
from .template import gemmi

if TYPE_CHECKING:
    from easyscience.Utils.typing import B


class SpaceGroup(CIF_Template):
    _CIF_SECTION_NAME: ClassVar[str] = '_space_group'
    _CIF_CONVERSIONS: ClassVar[List[Tuple[str, str]]] = [
        ('space_group_HM_name', '_name_H-M_ref'),
        ('setting', '_IT_coordinate_system_code'),
        ('symmetry_ops', "'_symop.operation_xyz'"),
    ]
    _CIF_ALTERNATES: ClassVar[List[Tuple[str, List[str]]]] = [
        [
            'space_group_HM_name',
            [
                '_name_H-M_full',
                '_IT_number',
                '_name_Hall',
                '_name_H-M_alt',
                '.name_H-M_full',
                '.IT_number',
                '.name_Hall',
                '.name_H-M_alt',
            ],
        ],
    ]

    def __init__(self, reference_class=_SpaceGroup):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        kwargs = {}
        for item in self._CIF_CONVERSIONS[0:2]:
            value = block.find_pair_item(self._CIF_SECTION_NAME + item[1])
            if value is None:
                a, b = zip(*self._CIF_ALTERNATES)
                if item[0] in a:
                    idx = a.index(item[0])
                    for i in b[idx]:
                        value = block.find_pair_item(self._CIF_SECTION_NAME + i)
                        if value is not None:
                            break
            if value is None:
                continue
            V, E, F = self.string_to_variable(value.pair[1])
            kwargs[item[0]] = V
        if not kwargs:
            item = self._CIF_CONVERSIONS[2]
            loop = list(block.find_values(self._CIF_SECTION_NAME + item[1]))
            ops = []
            for this_item in list(loop):
                ops.append(SymmOp.from_xyz_string(this_item))
            kwargs[item[0]] = ops
        return self._CIF_CLASS(**kwargs)

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        if not obj.is_custom:
            for item in self._CIF_CONVERSIONS[0:2]:
                value = getattr(obj, item[0])
                if value:
                    block.set_pair(self._CIF_SECTION_NAME + item[1], self.variable_to_string(value))
        else:
            loop = block.init_loop('_space_group_symop.', ['id', 'operation_xyz'])
            for i, op in enumerate(obj.symmetry_ops):
                loop.add_row([str(i + 1), "'" + op.as_xyz_string().replace(' ', '') + "'"])

    def from_cif_string(self, cif_string: str) -> List[B]:
        if 'data_' not in cif_string:
            cif_string = 'data_temp\n' + cif_string

        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import NoReturn
from typing import Tuple

from easycrystallography.Components.Lattice import Lattice as _Lattice

from .template import CIF_Template
from .template import gemmi

if TYPE_CHECKING:
    from easyscience.Utils.typing import B


class Lattice(CIF_Template):
    _CIF_SECTION_NAME: ClassVar[str] = '_cell'
    _CIF_CONVERSIONS: ClassVar[List[Tuple[str, str]]] = [
        ('length_a', '_length_a'),
        ('length_b', '_length_b'),
        ('length_c', '_length_c'),
        ('angle_alpha', '_angle_alpha'),
        ('angle_beta', '_angle_beta'),
        ('angle_gamma', '_angle_gamma'),
    ]

    def __init__(self, reference_class=_Lattice):
        super().__init__()
        self._CIF_CLASS = reference_class

    def from_cif_block(self, block: gemmi.cif.Block) -> B:
        kwargs = {}
        errors = {}
        is_fixed = {}
        for item in self._CIF_CONVERSIONS:
            value = block.find_pair_item(self._CIF_SECTION_NAME + item[1])
            if value is None:
                # try the new CIF format with '.' instead of '_'
                value = block.find_pair_item(self._CIF_SECTION_NAME + item[1].replace('_', '.', 1))
            V, E, F = self.string_to_variable(value.pair[1])
            if E:
                errors[item[0]] = E
            if F is not None and not F:
                is_fixed[item[0]] = F
            kwargs[item[0]] = V
        obj = self._CIF_CLASS(**kwargs)
        for error in errors.keys():
            setattr(getattr(obj, error), 'error', errors[error])
        for atr in is_fixed.keys():
            setattr(getattr(obj, atr), 'fixed', is_fixed[atr])
        return obj

    def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
        for item in self._CIF_CONVERSIONS:
            value = getattr(obj, item[0])
            block.set_pair(self._CIF_SECTION_NAME + item[1], self.variable_to_string(value))

    def from_cif_string(self, cif_string: str) -> List[B]:
        if 'data_' not in cif_string:
            cif_string = 'data_temp\n' + cif_string

        cif_blocks = gemmi.cif.read_string(cif_string)
        objs = []
        for block in cif_blocks:
            objs.append(self.from_cif_block(block))
        return objs



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from easycrystallography.Symmetry.SymOp import SymmOp


class Twin:
    def __init__(self, origin=(0, 0, 0), theta: float = 0.0, phi: float = 0.0):
        self.theta = theta
        self.phi = phi
        self.origin = list(origin)
        self.axis1 = [1, 0, 0]
        self.axis2 = [0, 0, 1]

    @property
    def operation(self):
        return SymmOp.from_origin_axis_angle(self.origin, self.axis1, self.phi) * SymmOp.from_origin_axis_angle(
            self.origin, self.axis2, self.theta
        )



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from typing import List

import pytest
from copy import deepcopy
import easyscience
import numpy as np
from easycrystallography.Components.Site import Site, PeriodicSite, Parameter, _SITE_DETAILS
from easyscience import global_object


site_details = [("Al", "Al"), ("Fe", "Fe3+"), ("TEST", "H")]


@pytest.fixture
def instance(request):
    def class_creation(*args, **kwargs):
        return Site(*request.param, *args, **kwargs)

    return class_creation


def _generate_inputs(do_occ=True):
    # These are the parameters which will always be present
    # These will be the optional parameters
    a = [1]
    if do_occ:
        a = [0, 0.1, 1.1]
    b = [0.1, 0.5]

    advanced = {
        "occupancy": a.copy(),
        "fract_x": b.copy(),
        "fract_y": b.copy(),
        "fract_z": b.copy(),
    }
    occ = _SITE_DETAILS["occupancy"].copy()
    occ["value"] = a.copy()

    pos = _SITE_DETAILS["position"].copy()
    pos["value"] = b.copy()

    advanced_result = {
        "occupancy": {"name": "occupancy", **occ},
        "fract_x": {"name": "fract_x", **pos},
        "fract_y": {"name": "fract_y", **pos},
        "fract_z": {"name": "fract_z", **pos},
    }

    def create_entry(base, key, value, ref, ref_key=None):
        this_temp = deepcopy(base)
        if this_temp:
            for item in base:
                test, res = item
                new_opt = deepcopy(test)
                new_res = deepcopy(res)
                if ref_key is None:
                    ref_key = key
                new_res[ref_key] = ref
                new_opt[key] = value
                this_temp.append((new_opt, new_res))
        else:
            this_temp.append(({key: value}, {ref_key: ref}))
        return this_temp

    temp = []
    for add_opt in advanced.keys():
        if isinstance(advanced[add_opt], list):
            for idx, item in enumerate(advanced[add_opt]):
                temp = create_entry(
                    temp,
                    add_opt,
                    item,
                    advanced_result[add_opt]["value"][idx],
                    ref_key=advanced_result[add_opt]["name"],
                )
        else:
            temp = create_entry(
                temp,
                add_opt,
                advanced[add_opt],
                advanced_result[add_opt]["value"],
                ref_key=advanced_result[add_opt]["name"],
            )
    return temp


@pytest.mark.parametrize("instance", site_details, indirect=True)
@pytest.mark.parametrize("element, expected", _generate_inputs())
def test_Site_creation(instance, element: List, expected: dict):
    global_object.map._clear()
    d = instance(**element)
    for field in expected.keys():
        ref = expected[field]
        obtained = getattr(d, field)
        if isinstance(obtained, Parameter):
            assert obtained.raw_value == ref
        else:
            assert obtained == ref


@pytest.mark.parametrize("label, elm", site_details)
def test_Site_default(label, elm):
    site = Site(label, elm)

    assert site.name == label
    assert site.specie.raw_value == elm

    positions = _SITE_DETAILS["position"]
    for item_label in ["fract_x", "fract_y", "fract_z"]:
        item = getattr(site, item_label)
        for key in positions.keys():
            test_key = key
            if key == "value":
                test_key = "raw_value"
            assert getattr(item, test_key) == positions[key]

    occupancy = _SITE_DETAILS["occupancy"]
    for item_label in ["occupancy"]:
        item = getattr(site, item_label)
        for key in occupancy.keys():
            test_key = key
            if key == "value":
                test_key = "raw_value"
            assert getattr(item, test_key) == occupancy[key]


# FAILED tests/unit_tests/Components/test_Site.py::test_Site_short_pos[element51-expected51-instance1] - RuntimeError: dictionary changed size during iteration
# FAILED tests/unit_tests/Components/test_Site.py::test_Site_fract_coords[element14-expected14-instance2] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
@pytest.mark.parametrize("instance", site_details, indirect=True)
@pytest.mark.parametrize("element, expected", _generate_inputs(do_occ=False))
def test_Site_short_pos(instance, element: List, expected: dict):
    global_object.map._clear()
    d = instance(**element)
    pars = [("x", "fract_x"), ("y", "fract_y"), ("z", "fract_z")]

    for par in pars:
        if par[1] in expected.keys():
            expected_value = expected[par[1]]
        else:
            expected_value = _SITE_DETAILS["position"]["value"]
        obtained = getattr(d, par[0])
        assert expected_value == obtained.raw_value


#FAILED tests/unit_tests/Components/test_Site.py::test_Site_fract_dist[second_pt0-element24-expected24-instance0] - RuntimeError: dictionary changed size during iteration
#FAILED tests/unit_tests/Components/test_Site.py::test_Site_fract_dist[second_pt0-element32-expected32-instance0] - RuntimeError: dictionary changed size during iteration
#FAILED tests/unit_tests/Components/test_Site.py::test_Site_fract_dist[second_pt1-element48-expected48-instance2] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
@pytest.mark.parametrize("instance", site_details, indirect=True)
@pytest.mark.parametrize("element, expected", _generate_inputs(do_occ=False))
def test_Site_fract_coords(instance, element: List, expected: dict):
    global_object.map._clear()
    d = instance(**element)

    el_for_check = ["fract_x", "fract_y", "fract_z"]

    expected_value = []
    for par in el_for_check:
        if par in expected.keys():
            expected_value.append(expected[par])
        else:
            expected_value.append(_SITE_DETAILS["position"]["value"])
    assert np.all(d.fract_coords == expected_value)


@pytest.mark.parametrize("instance", site_details, indirect=True)
@pytest.mark.parametrize("element, expected", _generate_inputs(do_occ=False))
@pytest.mark.parametrize(
    "second_pt", ([0.0, 0.0, 0.0], [0.25, 0.1, 0.1], [1 / 8, 1 / 3, 1 / 4])
)
def test_Site_fract_dist(instance, element: List, expected: dict, second_pt):
    global_object.map._clear()
    d = instance(**element)
    other_site = Site("H", "H", 1, *second_pt)
    expected_value = np.linalg.norm(np.array(second_pt) - d.fract_coords)
    assert np.all(d.fract_distance(other_site) == expected_value)


@pytest.mark.parametrize("label, elm", site_details)
@pytest.mark.parametrize("element, expected", _generate_inputs(do_occ=False))
def test_Site_repr(label, elm, element, expected):
    global_object.map._clear()
    d = Site(label, elm, **element)

    el_for_check = ["fract_x", "fract_y", "fract_z"]

    expected_value = []
    for par in el_for_check:
        if par in expected.keys():
            expected_value.append(expected[par])
        else:
            expected_value.append(float(_SITE_DETAILS["position"]["value"]))

    assert (
        str(d)
        == f"Atom {label} ({elm}) @ ({expected_value[0]}, {expected_value[1]}, {expected_value[2]})"
    )


@pytest.mark.parametrize("label, elm", site_details)
def test_Site_as_dict(label, elm):
    global_object.map._clear()
    s = Site(label, elm)
    obtained = s.as_dict()
    expected = {
        "@module": "easycrystallography.Components.Site",
        "@class": "Site",
        "unique_name": None,
        "label": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Descriptor",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "label",
            "value": label,
            "units": "dimensionless",
            "description": "A unique identifier for a particular site in the crystal",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "display_name": "label",
            "enabled": True,
        },
        # Note that we are skipping specie checking as it it covered in another file...
        # 'specie': {
        #      '@module':  'easyCore.Elements.Basic.Specie', '@class': 'Specie',
        #      'unique_name':      None,
        #      'specie':
        #                  {
        #                      '@module':         'easyCore.Elements.periodic_table',
        #                      '@class':          'Specie' + ('s' if '+' in elm else ''),
        #                      'element':         elm,
        #                      'oxidation_state': 0.0
        #                  },
        #      'value':    elm,
        #      'units':    'dimensionless',
        #  },
        "occupancy": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "occupancy",
            "value": 1.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "The fraction of the atom type present at this site.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_x": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_x",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_y": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_y",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_z": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_z",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "interface": None,
    }

    def check_dict(check, item):
        if isinstance(check, dict) and isinstance(item, dict):
            for this_check_key in check.keys():
                if this_check_key == "unique_name":
                    continue
                check_dict(check[this_check_key], item[this_check_key])
        else:
            assert isinstance(item, type(check))
            assert item == check

    check_dict(expected, obtained)


@pytest.mark.parametrize("label, elm", site_details)
def test_Site_from_dict(label, elm):
    d = {
        "@module": "easycrystallography.Components.Site",
        "@class": "Site",
        "unique_name": None,
        "label": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Descriptor",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "label",
            "value": label,
            "units": "dimensionless",
            "description": "A unique identifier for a particular site in the crystal",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "display_name": "label",
            "enabled": True,
        },
        "specie": {
            "@module": "easycrystallography.Components.Specie",
            "@class": "Specie",
            "@version": easyscience.__version__,
            "unique_name": None,
            "specie": elm,
            "value": elm,
            "units": "dimensionless",
        },
        "occupancy": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "occupancy",
            "value": 1.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "The fraction of the atom type present at this site.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_x": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_x",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_y": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_y",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "fract_z": {
            "@module": "easyscience.Objects.Variable",
            "@class": "Parameter",
            "@version": easyscience.__version__,
            "unique_name": None,
            "name": "fract_z",
            "value": 0.0,
            "error": 0.0,
            "min": -np.inf,
            "max": np.inf,
            "fixed": True,
            "description": "Atom-site coordinate as fractions of the unit cell length.",
            "url": "https://docs.easydiffraction.org/lib/dictionaries/_atom_site/",
            "units": "dimensionless",
            "enabled": True,
        },
        "interface": None,
    }

    s = Site.from_dict(d)

    def check_dict(check, item):
        if isinstance(check, dict) and isinstance(item, dict):
            for this_check_key in check.keys():
                if this_check_key == "unique_name" or this_check_key == "specie":
                    continue
                check_dict(check[this_check_key], item[this_check_key])
        else:
            assert isinstance(item, type(check))
            assert item == check

    check_dict(d, s.as_dict())


def test_site_with_adp():
    from easycrystallography.Components.AtomicDisplacement import AtomicDisplacement
    adp = AtomicDisplacement()
    site = Site(adp=adp)
    assert hasattr(site, "adp")
    assert site.adp is adp


# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

import pytest
import itertools
import numpy as np

from easyscience.Objects.ObjectClasses import Descriptor, Parameter
from easyscience import global_object
from easycrystallography.Components.SpaceGroup import SpaceGroup, SG_DETAILS as _SG_DETAILS
from easycrystallography.Symmetry.groups import SpaceGroup as SG

SG_DETAILS = _SG_DETAILS.copy()
del SG_DETAILS['symmetry_ops']

known_conversions = {
    "A e m 2":  'A b m 2',
    "B m e 2":  'B m a 2',
    "B 2 e m":  'B 2 c m',
    "C 2 m e":  'C 2 m b',
    "C m 2 e":  'C m 2 a',
    "A e 2 m":  'A c 2 m',
    "A e a 2":  'A b a 2',
    "B b e 2":  'B b a 2',
    "B 2 e b":  'B 2 c b',
    "C 2 c e":  'C 2 c b',
    "C c 2 e":  'C c 2 a',
    "A e 2 a":  'A c 2 a',
    "C m c e":  'C m c a',
    "C c m e":  'C c m b',
    "A e m a":  'A b m a',
    "A e a m":  'A c a m',
    "B b e m":  'B b c m',
    "B m e b":  'B m a b',
    "C m m e":  'C m m a',
    "A e m m":  'A b m m',
    "B m e m":  'B m c m',
    "C c c e":  'C c c a',
    "A e a a":  'A b a a',
    "B b e b":  'B b c b',
    'B 1 21 1': 'B 1 1 2',

}

SYM = [value for value in SG.SYMM_OPS
       if '(' not in value['hermann_mauguin_fmt']
       or '(' not in value['hermann_mauguin']
       or '(' not in value['universal_h_m']]


def test_SpaceGroup_fromDescriptor():
    sg_items = ['space_group_HM_name', 'P 1']

    d = Descriptor(*sg_items)
    sg = SpaceGroup(d)
    assert sg.space_group_HM_name.raw_value == 'P 1'

    with pytest.raises(ValueError):
        p = Parameter('space_group_HM_name', 1)
        sg = SpaceGroup(p)


def test_SpaceGroup_default():
    sg = SpaceGroup()
    assert sg.setting is None

    for selector in SG_DETAILS.keys():
        f = getattr(sg, selector)
        if f is None:
            continue
        for item in SG_DETAILS[selector].keys():
            g_item = item
            if item == 'value':
                g_item = 'raw_value'
            assert getattr(f, g_item) == SG_DETAILS[selector][item]

    assert isinstance(sg.space_group_HM_name, Descriptor)


@pytest.mark.parametrize('sg_in', [sg['hermann_mauguin_fmt'] for sg in SYM])
def test_SpaceGroup_fromPars_HM_Full(sg_in):
    if sg_in in ['C 2 e b', 'R 1 2/c 1 ("rhombohedral" setting)', 'B 1 21/m 1', 'B 1 21 1']:
        return  # This is a known issue

    sg_p = SpaceGroup(sg_in)

    for selector in SG_DETAILS.keys():
        f = getattr(sg_p, selector)
        if f is None:
            continue
        for item in SG_DETAILS[selector].keys():
            g_item = item
            f_value = SG_DETAILS[selector][item]
            if item == 'value':
                g_item = 'raw_value'
                f_value = sg_in.split(':')
            if f_value[0] in known_conversions.keys():
                f_value[0] = known_conversions[f_value[0]]
            assert getattr(f, g_item) in f_value

# FAILED tests/unit_tests/Components/test_SpaceGroup.py::test_SpaceGroup_fromPars_HM_noSpace[I41/a:2] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
@pytest.mark.parametrize('sg_in', SYM, ids=[sg['hermann_mauguin'] for sg in SYM])
def test_SpaceGroup_fromPars_HM_noSpace(sg_in):
    global_object.map._clear()
    if sg_in['hermann_mauguin'] in ['C2eb', 'R12/c1("rhombohedral"setting)', 'B1211', 'B121/m1', 'P4bm',
                                    'C1c1', 'Pmc21', 'Cmm2', 'P121/c1', 'Pmma', 'P12/c1', 'Pmmm', 'P1211', 'Pnma']:
        return  # This is a known issue

    sg_p = SpaceGroup(sg_in['hermann_mauguin'])

    for selector in SG_DETAILS.keys():
        f = getattr(sg_p, selector)
        if f is None:
            continue
        for item in SG_DETAILS[selector].keys():
            g_item = item
            f_value = SG_DETAILS[selector][item]
            if item == 'value':
                g_item = 'raw_value'
                f_value = sg_in['hermann_mauguin_fmt'].split(':')
            if f_value[0] in known_conversions.keys():
                f_value[0] = known_conversions[f_value[0]]
            assert getattr(f, g_item) in f_value

# FAILED tests/unit_tests/Components/test_SpaceGroup.py::test_SpaceGroup_fromPars_HM_noSpace[P-42c] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
@pytest.mark.parametrize('sg_in', SYM, ids=[sg['universal_h_m'] for sg in SYM])
def test_SpaceGroup_fromPars_HM_noSpace(sg_in):
    global_object.map._clear()
    if sg_in['hermann_mauguin'] in ['C2eb', 'R12/c1("rhombohedral"setting)', 'B1211', 'B121/m1', 'P4bm',
                                    'C1c1', 'Pmc21', 'Cmm2', 'P121/c1', 'Pmma', 'P12/c1', 'Pmmm', 'P1211', 'Pnma']:
        return  # This is a known issue

    sg_p = SpaceGroup(sg_in['universal_h_m'])

    for selector in SG_DETAILS.keys():
        f = getattr(sg_p, selector)
        if f is None:
            continue
        for item in SG_DETAILS[selector].keys():
            g_item = item
            f_value = SG_DETAILS[selector][item]
            if item == 'value':
                g_item = 'raw_value'
                f_value = sg_in['hermann_mauguin_fmt'].split(':')
            if f_value[0] in known_conversions.keys():
                f_value[0] = known_conversions[f_value[0]]
            assert getattr(f, g_item) in f_value


@pytest.mark.parametrize('sg_int', range(1, 231), ids=[f'spacegroup_int_{s_id}' for s_id in range(1, 231)])
def test_SpaceGroup_fromIntNumber(sg_int: int):
    sg_p = SpaceGroup.from_int_number(sg_int)

    for selector in SG_DETAILS.keys():
        f = getattr(sg_p, selector)
        if f is None:
            continue
        for item in SG_DETAILS[selector].keys():
            g_item = item
            f_value = SG_DETAILS[selector][item]
            if item == 'value':
                g_item = 'raw_value'
                for opt in SG.SYMM_OPS:
                    if opt['number'] == sg_int:
                        f_value = opt['hermann_mauguin_fmt'].split(':')
                        break
            if f_value[0] in known_conversions.keys():
                f_value[0] = known_conversions[f_value[0]]
            assert getattr(f, g_item) in f_value


@pytest.mark.parametrize('sg_int,setting', itertools.product([146, 148, 155, 160, 161, 166, 167], [True, False]))
def test_SpaceGroup_fromIntNumber_HexTest(sg_int: int, setting: bool):
    sg_p = SpaceGroup.from_int_number(sg_int, setting)

    for selector in SG_DETAILS.keys():
        f = getattr(sg_p, selector)
        for item in SG_DETAILS[selector].keys():
            g_item = item
            f_value = SG_DETAILS[selector][item]
            if item == 'value':
                g_item = 'raw_value'
                for opt in SG.SYMM_OPS:
                    if opt['number'] == sg_int:
                        f_value: str = opt['hermann_mauguin_fmt']
                        if f_value.endswith(':H') and setting or f_value.endswith(':R') and not setting:
                            break
                assert getattr(f, g_item) in f_value.split(':')
                return
        assert getattr(f, g_item) in f_value


def test_SpaceGroup_as_dict():
    sg_p = SpaceGroup.from_int_number(146)
    sg_p_dict = sg_p.as_dict()
    del sg_p_dict['setting']['unique_name']
    del sg_p_dict['space_group_HM_name']['unique_name']
    del sg_p_dict['unique_name']
    del sg_p_dict['space_group_HM_name']['@class']
    del sg_p_dict['setting']['@class']
    del sg_p_dict['space_group_HM_name']['@module']
    del sg_p_dict['setting']['@module']
    del sg_p_dict['space_group_HM_name']['@version']
    del sg_p_dict['setting']['@version']
    del sg_p_dict['@class']
    del sg_p_dict['@module']
    del sg_p_dict['@version']

    assert sg_p_dict == {
                                   'symmetry_ops':        None,
                                   'setting':             {
                                       'units':        'dimensionless',
                                       'enabled':      True,
                                       'name':         'coordinate-code',
                                       'display_name': 'coordinate-code',
                                       'url': 'https://docs.easydiffraction.org/lib/dictionaries/_space_group/',
                                       'description':  'A qualifier taken from the enumeration list identifying which '
                                                       'setting in International Tables for Crystallography Volume A '
                                                       '(2002) (IT) is used.',
                                       'value':        'H'
                                   },
                                   'interface':           None,
                                   'space_group_HM_name': {
                                       'units':        'dimensionless',
                                       'enabled':      True,
                                       'name':         'hermann_mauguin',
                                       'display_name': 'hermann_mauguin',
                                       'url':          'https://docs.easydiffraction.org/lib/dictionaries/_space_group/',
                                       'description':  'Hermann-Mauguin symbols given in Table 4.3.2.1 of '
                                                       'International Tables for Crystallography Vol. A (2002) or a '
                                                       'Hermann-Mauguin symbol for a conventional or unconventional '
                                                       'setting.',
                                       'value':        'R 3'}}


def test_SpaceGroup_change_setting():
    sg_p = SpaceGroup('P b a n')
    assert sg_p.setting_str == '1'
    old_ops = sg_p.symmetry_ops
    sg_p.setting = '2'
    assert sg_p.setting.raw_value == '2'
    assert not np.all(sg_p.symmetry_ops == old_ops)

# DISABLE UNTIL UNIQUE_NAME IS FIXED

#def test_SpaceGroup_from_dict():
#    from time import sleep
#    sg_p = SpaceGroup.from_int_number(146)
#    d = sg_p.as_dict()
#    del d['setting']['unique_name']
#    del d['space_group_HM_name']['unique_name']
#    global_object.map._clear()
#    sg_2 = SpaceGroup.from_dict(d)
#    # temporarily disabled due to global_object acting up
#    # TODO: enable the test once map._clear() behaves
#    #sleep(5)
#    #assert sg_2.as_data_dict() == sg_p.as_data_dict()

# def test_SpaceGroup_to_cif_str():
#     sg_p = SpaceGroup.from_int_number(15)
#     s = sg_p.to_cif_str()
#     assert s == '_space_group_name_H-M_ref C 1 2/c 1\n'


# def test_SpaceGroup_to_cif_str_with_setting():
#     sg_p = SpaceGroup.from_int_number(146)
#     s = sg_p.to_cif_str()
#     assert s == '_space_group_name_H-M_ref R 3\n_space_group_IT_coordinate_system_code H\n'


def testSpaceGroup_from_SymOps():
    from easycrystallography.Symmetry.SymOp import SymmOp
    ops_str = 'x, y, z;-x, y, -z+1/2;-x, -y, -z;x, -y, z+1/2;x+1/2, y+1/2, z;-x+1/2, y+1/2, -z+1/2;-x+1/2, -y+1/2, ' \
              '-z;x+1/2, -y+1/2, z+1/2'
    ops_str_list = ops_str.split(';')
    ops = [SymmOp.from_xyz_string(op) for op in ops_str_list]
    spg = SpaceGroup.from_symOps(ops)
    assert spg.int_number == 15


def testSpaceGroup_from_xyz_string():
    ops_str = 'x, y, z;-x, y, -z+1/2;-x, -y, -z;x, -y, z+1/2;x+1/2, y+1/2, z;-x+1/2, y+1/2, -z+1/2;-x+1/2, -y+1/2, ' \
              '-z;x+1/2, -y+1/2, z+1/2'
    ops_str_list = ops_str.split(';')

    spg = SpaceGroup.from_xyz_string(ops_str)
    assert spg.int_number == 15
    spg = SpaceGroup.from_xyz_string(ops_str_list)
    assert spg.int_number == 15


def testSpaceGroup_from_SymOps():
    from easycrystallography.Symmetry.SymOp import SymmOp
    ops_str = 'x, y, z;-x, y, -z+1/2;-x, -y, -z;x, -y, z+1/2;x+1/2, y+1/2, z;-x+1/2, y+1/2, -z+1/2;-x+1/2, -y+1/2, ' \
              '-z;x+1/2, -y+1/2, z+1/2'
    ops_str_list = ops_str.split(';')
    ops = [SymmOp.from_xyz_string(op) for op in ops_str_list]
    sym_mat_rot = [op.rotation_matrix for op in ops]
    sym_mat_trans = [op.translation_vector for op in ops]
    spg = SpaceGroup.from_symMatrices(sym_mat_rot, sym_mat_trans)
    assert spg.int_number == 15


def testSpaceGroup_from_gemmi_ops():
    from gemmi import Op, GroupOps
    ops_str = 'x, y, z;-x, y, -z+1/2;-x, -y, -z;x, -y, z+1/2;x+1/2, y+1/2, z;-x+1/2, y+1/2, -z+1/2;-x+1/2, -y+1/2, ' \
              '-z;x+1/2, -y+1/2, z+1/2'
    ops_str_list = ops_str.split(';')
    ops = [Op(op) for op in ops_str_list]

    spg = SpaceGroup.from_gemmi_operations(ops)
    assert spg.int_number == 15

    spg = SpaceGroup.from_gemmi_operations(GroupOps(ops))
    assert spg.int_number == 15



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import Optional
from typing import TypeVar
from typing import Union

import numpy as np
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Descriptor
from easyscience.Objects.ObjectClasses import Parameter
from easyscience.Utils.classTools import addProp
from easyscience.Utils.classTools import removeProp

if TYPE_CHECKING:
    from easyscience.Utils.typing import iF

_ANIO_DETAILS = {
    'adp_type': {
        'description': 'A standard code used to describe the type of atomic displacement parameters used for the site.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 'Uani',
    },
    'Uani': {
        'description': 'Isotropic atomic displacement parameter, or equivalent isotropic atomic  displacement '
        'parameter, U(equiv), in angstroms squared, calculated from anisotropic atomic displacement  '
        'parameters.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 0.0,
        'units': 'angstrom^2',
        'fixed': True,
    },
    'Uiso': {
        'description': 'The standard anisotropic atomic displacement components in angstroms squared which appear in '
        'the structure-factor term.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 0.0,
        'min': 0,
        'max': np.inf,
        'units': 'angstrom^2',
        'fixed': True,
    },
    'Bani': {
        'description': 'The standard anisotropic atomic displacement components in angstroms squared which appear in '
        'the structure-factor term.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 0.0,
        'units': 'angstrom^2',
        'fixed': True,
    },
    'Biso': {
        'description': 'Isotropic atomic displacement parameter, or equivalent isotropic atomic displacement '
        'parameter, B(equiv), in angstroms squared, calculated from anisotropic displacement '
        'components.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_atom_site/',
        'value': 0.0,
        'min': 0,
        'max': np.inf,
        'units': 'angstrom^2',
        'fixed': True,
    },
}


class AdpBase(BaseObj):
    def __init__(self, *args, **kwargs):
        super(AdpBase, self).__init__(*args, **kwargs)

    @property
    def matrix(self) -> np.ndarray:
        matrix = np.zeros([3, 3])
        pars = self.get_parameters()
        if len(pars) == 1:
            np.fill_diagonal(matrix, pars[0].raw_value)
        elif len(pars) == 6:
            matrix[0, 0] = pars[0].raw_value
            matrix[0, 1] = pars[1].raw_value
            matrix[0, 2] = pars[2].raw_value
            matrix[1, 1] = pars[3].raw_value
            matrix[1, 2] = pars[4].raw_value
            matrix[2, 2] = pars[5].raw_value
        return matrix

    def __repr__(self):
        s = f'{self.name} - ('
        for par in self.get_parameters():
            s += f'{par.name}: {par.raw_value}, '
        s = s[:-2] + ')'
        return s


class Anisotropic(AdpBase):
    U_11: ClassVar[Parameter]
    U_12: ClassVar[Parameter]
    U_13: ClassVar[Parameter]
    U_22: ClassVar[Parameter]
    U_23: ClassVar[Parameter]
    U_33: ClassVar[Parameter]

    def __init__(
        self,
        Uiso_ani: Optional[float] = None,
        U_11: Optional[Union[Parameter, float]] = None,
        U_12: Optional[Union[Parameter, float]] = None,
        U_13: Optional[Union[Parameter, float]] = None,
        U_22: Optional[Union[Parameter, float]] = None,
        U_23: Optional[Union[Parameter, float]] = None,
        U_33: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
    ):
        super(Anisotropic, self).__init__(
            'anisoU',
            U_11=Parameter('U_11', **_ANIO_DETAILS['Uani']),
            U_12=Parameter('U_12', **_ANIO_DETAILS['Uani']),
            U_13=Parameter('U_13', **_ANIO_DETAILS['Uani']),
            U_22=Parameter('U_22', **_ANIO_DETAILS['Uani']),
            U_23=Parameter('U_23', **_ANIO_DETAILS['Uani']),
            U_33=Parameter('U_33', **_ANIO_DETAILS['Uani']),
        )
        if U_11 is not None:
            self.U_11 = U_11
        if U_12 is not None:
            self.U_12 = U_12
        if U_13 is not None:
            self.U_13 = U_13
        if U_22 is not None:
            self.U_22 = U_22
        if U_23 is not None:
            self.U_23 = U_23
        if U_33 is not None:
            self.U_33 = U_33
        if Uiso_ani is not None:
            self.Uiso_ani = Uiso_ani
        else:
            # for cubic, tetragonal, and orthorhombic systems
            self.Uiso_ani = (self.U_11.raw_value + self.U_22.raw_value + self.U_33.raw_value) / 3.0
        self.interface = interface


class Isotropic(AdpBase):
    Uiso: ClassVar[Parameter]

    def __init__(
        self,
        Uiso: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
    ):
        super(Isotropic, self).__init__('Uiso', Uiso=Parameter('Uiso', **_ANIO_DETAILS['Uiso']))
        if Uiso is not None:
            self.Uiso = Uiso
        self.interface = interface


class AnisotropicBij(AdpBase):
    B_11: ClassVar[Parameter]
    B_12: ClassVar[Parameter]
    B_13: ClassVar[Parameter]
    B_22: ClassVar[Parameter]
    B_23: ClassVar[Parameter]
    B_33: ClassVar[Parameter]

    def __init__(
        self,
        B_11: Optional[Union[Parameter, float]] = None,
        B_12: Optional[Union[Parameter, float]] = None,
        B_13: Optional[Union[Parameter, float]] = None,
        B_22: Optional[Union[Parameter, float]] = None,
        B_23: Optional[Union[Parameter, float]] = None,
        B_33: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
    ):
        super(AnisotropicBij, self).__init__(
            'anisoB',
            **{name: Parameter(name, **_ANIO_DETAILS['Bani']) for name in ['B_11', 'B_12', 'B_13', 'B_22', 'B_23', 'B_33']},
        )
        if B_11 is not None:
            self.B_11 = B_11
        if B_12 is not None:
            self.B_12 = B_12
        if B_13 is not None:
            self.B_13 = B_13
        if B_22 is not None:
            self.B_22 = B_22
        if B_23 is not None:
            self.B_23 = B_23
        if B_33 is not None:
            self.B_33 = B_33
        self.interface = interface


class IsotropicB(AdpBase):
    Biso: ClassVar[Parameter]

    def __init__(
        self,
        Biso: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
    ):
        super(IsotropicB, self).__init__('Biso', Biso=Parameter('Biso', **_ANIO_DETAILS['Biso']))
        if Biso is not None:
            self.Biso = Biso
        self.interface = interface


_AVAILABLE_ISO_TYPES = {
    'Uani': Anisotropic,
    'Uiso': Isotropic,
    # 'Uovl': 'Overall',
    # 'Umpe': 'MultipoleExpansion',
    'Bani': AnisotropicBij,
    'Biso': IsotropicB,
    # 'Bovl': 'OverallB'
}

if TYPE_CHECKING:
    AB = TypeVar('AB', bound=AdpBase)


class AtomicDisplacement(BaseObj):
    adp_type: ClassVar[Descriptor]
    adp_class: ClassVar[AB]

    def __init__(
        self,
        adp_type: Optional[Union[Descriptor, str]] = None,
        interface: Optional[iF] = None,
        **kwargs,
    ):
        if adp_type is None:
            adp_type = 'Uiso'
        if isinstance(adp_type, str):
            adp_type = Descriptor('adp_type', adp_type)
        adp_class_name = adp_type.raw_value
        if adp_class_name in _AVAILABLE_ISO_TYPES.keys():
            adp_class = _AVAILABLE_ISO_TYPES[adp_class_name]
            # enable passing ADP parameters to constructor
            if 'adp_class' in kwargs.keys():
                m = getattr(kwargs['adp_class'], adp_class_name)
                kwargs[adp_class_name] = m
                _ = kwargs.pop('adp_class')
            adp = adp_class(**kwargs, interface=interface)
        else:
            raise AttributeError(f'{adp_class_name} is not a valid adp type')
        super(AtomicDisplacement, self).__init__('adp', adp_type=adp_type, adp_class=adp)
        for par in adp.get_parameters():
            addProp(
                self,
                par.name,
                fget=self.__a_getter(par.name),
                fset=self.__a_setter(par.name),
            )
        self.interface = interface

    def switch_type(self, adp_string: str, **kwargs):
        # if adp_string in _AVAILABLE_ISO_TYPES.keys():
        # adp_class = _AVAILABLE_ISO_TYPES[adp_string]
        # if kwargs:
        #     adp = adp_class(**kwargs, interface=self.interface)
        # else:
        #     adp = adp_class(interface=self.interface)
        # else:
        #     raise AttributeError(f"{adp_string} is not a valid adp type")
        for par in self.adp_class.get_parameters():
            removeProp(self, par.name)
        self.__init__(adp_type=adp_string, **kwargs)

    @property
    def available_types(self) -> List[str]:
        return [name for name in _AVAILABLE_ISO_TYPES.keys()]

    @staticmethod
    def __a_getter(key: str):
        def getter(obj):
            return obj.adp_class._kwargs[key]

        return getter

    @staticmethod
    def __a_setter(key):
        def setter(obj, value):
            obj.adp_class._kwargs[key].value = value

        return setter



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

# from __future__ import annotations
#
# from typing import List, NoReturn, TYPE_CHECKING, ClassVar, Tuple, Dict
#
# from easyscience.Objects.ObjectClasses import B
# from easycrystallography.Structures.Phase import Phase as _Phase, Phases as _Phases
# from .template import CIF_Template, gemmi
# from . import *
#
#
# class Phase(CIF_Template):
#
#     def __init__(self, reference_class=_Phase):
#         super().__init__()
#         self._CIF_CLASS = reference_class
#
#     def from_cif_block(self, block: gemmi.cif.Block) -> B:
#         pass
#
#     def add_to_cif_block(self, obj: B, block: gemmi.cif.Block) -> NoReturn:
#         pass
#
#     def from_cif_string(self, cif_string: str) -> List[B]:
#         pass
#



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

import re
import textwrap
import warnings
from collections import OrderedDict
from collections import deque
from math import floor
from math import log10
from numbers import Number
from typing import List

# STANDARD CIF
# _MAX_LEN = 70
# _MAX_LABEL_LEN = 65

_MAX_LEN = 140
_MAX_LABEL_LEN = 130


class FakeItem:
    def __init__(self, value: float, error=None, fixed: bool = None):
        self.raw_value = value
        if fixed is not None:
            self.fixed = fixed
            self.error = 0
        if error is not None:
            self.error = error


class FakeCore:
    def __init__(self):
        self._kwargs = {}


class ItemHolder:
    def __init__(self, item, decimal_places: int = 8):
        self.maxlen = _MAX_LEN

        self.value = item.raw_value
        self.fixed = None
        self.error = None
        self.decimal_places = decimal_places
        if hasattr(item, 'fixed'):
            self.fixed = item.fixed
            if item.error != 0:
                self.error = item.error

    def _get_error_digits(self) -> int:
        return len(f'{round(self.error, self.decimal_places)}'.split('.')[-1])

    def __str__(self) -> str:
        if isinstance(self.value, Number):
            initial_str = '{:.' + str(self.decimal_places) + 'f}'
            s = initial_str.format(round(self.value, self.decimal_places))
            if self.error is not None:
                # x_exp = int(floor(log10(self.value)))
                xe_exp = int(floor(log10(self.error)))

                # uncertainty
                un_exp = xe_exp - self.decimal_places + 1
                un_int = round(self.error * 10 ** (-un_exp))

                # nominal value
                no_exp = un_exp
                no_int = round(self.value * 10 ** (-no_exp))

                # format - nom(unc)
                fmt = '%%.%df' % max(0, -no_exp)
                s = (fmt + '(%.0f)') % (
                    no_int * 10**no_exp,
                    un_int * 10 ** max(0, un_exp),
                )
        elif isinstance(self.value, str):
            s = '{:s}'.format(self.value)
        else:
            s = '{:s}'.format(str(self.value))
        # THIS IS THE OLD CODE, KEPT FOR REFERENCE
        # s = "{}"
        # if isinstance(self.value, str):
        #     s = "{:s}".format(self.value)
        # else:
        #     v_in = [round(self.value, self.decimal_places)]
        #     if self.error is not None:
        #         digits = self._get_error_digits()
        #         if digits > self.decimal_places:
        #             v_in = [round(self.value, digits)]
        #         this_err = int(self.error * 10 ** digits)
        #         v_in.append(this_err)
        #         digits = digits - (len(str(this_err)) - 1)
        #         s = "{" + f":0.0{digits}f" + "}({})"
        #     s = s.format(*v_in)
        if self.fixed is not None and not self.fixed and self.error is None:
            s += '()'
        return self._format_field(s)

    def _format_field(self, v):
        v = v.__str__().strip()
        if len(v) > self.maxlen:
            return ';\n' + textwrap.fill(v, self.maxlen) + '\n;'
        # add quotes if necessary
        if v == '':
            return '""'
        if (' ' in v or v[0] == '_') and not (v[0] == "'" and v[-1] == "'") and not (v[0] == '"' and v[-1] == '"'):
            if "'" in v:
                q = '"'
            else:
                q = "'"
            v = q + v + q
        return v

    @staticmethod
    def _makeFakeItem(in_string: str) -> FakeItem:
        in_string = in_string.strip()
        if "'" in in_string:
            in_string = in_string.replace("'", '')
        fixed = None
        error = None
        tokens = in_string.split('(')
        try:
            value = float(tokens[0])
        except ValueError:
            value = tokens[0]
            return FakeItem(value, error, fixed)
        if len(tokens) > 1:
            fixed = False
            if tokens[1][0] != ')':
                error = (10 ** -(len(f'{tokens[0]}'.split('.')[1]) + len(tokens[1][:-1]) - 1)) * int(tokens[1][:-1])
        return FakeItem(value, error, fixed)

    @classmethod
    def from_string(cls, in_string: str):
        return cls(cls._makeFakeItem(in_string))

    def to_fake_item(self):
        return FakeItem(self.value, self.error, self.fixed)


class StarEntry(ItemHolder):
    def __init__(self, item, entry_name: str = None, prefix='_'):
        if entry_name is None:
            entry_name = item.name

        if len(entry_name) > _MAX_LABEL_LEN:
            raise AttributeError(f'Max label length is {int(_MAX_LABEL_LEN)}')

        self.name = entry_name
        self.prefix = prefix
        super(StarEntry, self).__init__(item)

    def __str__(self) -> str:
        s = '{}{}   '.format(self.prefix, self.name) + super(StarEntry, self).__str__()
        return s

    @classmethod
    def from_string(cls, input_str: str, name_conversion: str = None, prefix='_'):
        name, value = input_str.split('   ')
        name = name[len(prefix) :]
        if name_conversion:
            name = name_conversion
        return cls(cls._makeFakeItem(value), name)

    def to_class(self, cls, name_conversion: str = None):
        if name_conversion is None:
            name_conversion = self.name
        if hasattr(cls, 'from_pars'):
            new_obj = cls.from_pars(**{name_conversion: self.value})
        if hasattr(new_obj, 'fixed'):
            if self.fixed is not None:
                new_obj.fixed = self.fixed
        if hasattr(new_obj, 'error'):
            if self.error is not None:
                new_obj.error = self.error
        return new_obj


class StarProcess:
    @classmethod
    def _process_string(cls, string):
        # remove comments
        string = re.sub(r'(\s|^)#.*$', '', string, flags=re.MULTILINE)
        # remove empty lines
        string = re.sub(r'^\s*\n', '', string, flags=re.MULTILINE)
        # remove non_ascii
        # string = remove_non_ascii(string)
        # since line breaks in .cif files are mostly meaningless,
        # break up into a stream of tokens to parse, rejoining multiline
        # strings (between semicolons)
        q = deque()
        multiline = False
        ml = []
        # this regex splits on spaces, except when in quotes.
        # starting quotes must not be preceded by non-whitespace
        # (these get eaten by the first expression)
        # ending quotes must not be followed by non-whitespace
        p = re.compile(r"""([^'"\s][\S]*)|'(.*?)'(?!\S)|"(.*?)"(?!\S)""")
        for line in string.splitlines():
            if multiline:
                if line.startswith(';'):
                    multiline = False
                    q.append(('', '', '', ' '.join(ml)))
                    ml = []
                    line = line[1:].strip()
                else:
                    ml.append(line)
                    continue
            if line.startswith(';'):
                multiline = True
                ml.append(line[1:].strip())
            else:
                for s in p.findall(line):
                    # s is tuple. location of the data in the tuple
                    # depends on whether it was quoted in the input
                    q.append(s)
        return q

    @classmethod
    def _loadBlock(cls, in_string: str, prefix='_'):
        q = cls._process_string(in_string)
        data = OrderedDict()
        loops = []
        while q:
            s = q.popleft()
            # cif keys aren't in quotes, so show up in s[0]
            if s[0] == '_eof':
                break
            if s[0].startswith(prefix):
                try:
                    data[s[0]] = ''.join(q.popleft())
                except IndexError:
                    data[s[0]] = ''
            elif s[0].startswith('loop_'):
                columns = []
                items = []
                this_data = OrderedDict()
                while q:
                    s = q[0]
                    if s[0].startswith('loop_') or not s[0].startswith(prefix):
                        break
                    columns.append(''.join(q.popleft()))
                    this_data[columns[-1]] = []
                while q:
                    s = q[0]
                    if s[0].startswith('loop_') or s[0].startswith(prefix):
                        break
                    items.append(''.join(q.popleft()))
                n = len(items) // len(columns)
                if len(items) % n != 0:
                    raise ValueError('Wrong dimensions in loop_ block')
                # loops.append(columns)
                for k, v in zip(columns * n, items):
                    this_data[k].append(v.strip())
                loops.append(this_data)
            elif ''.join(s).strip() != '':
                warnings.warn('Possible issue in cif file' ' at line: {}'.format(''.join(s).strip()))
        return data, loops


class StarBase(StarProcess):
    def __init__(
        self,
        core,
        entry_names: List[str] = None,
        exclude: list = None,
        prefix: str = '_',
    ):
        if not hasattr(core, '__len__'):
            core = [core]
        self.data = core
        if exclude is None:
            exclude = []
        self.exclude = exclude

        if self.data:
            if entry_names is None:
                entry_names = [self.data[0]._kwargs[key].name for key in self.data[0]._kwargs.keys() if key not in exclude]
        else:
            entry_names = []

        self.prefix = prefix
        self.labels = entry_names
        self.maxlen = _MAX_LEN


class StarCollection(StarProcess):
    def __init__(self, *star_objects):
        self.data = star_objects

    def __str__(self) -> str:
        return '\n\n'.join([str(data) for data in self.data])

    @classmethod
    def from_string(cls, in_string, prefix='_'):
        in_string = '\n'.join([item for item in in_string.split('\n') if item and item[0] != '#'])

        blocks = in_string.split('data_')
        data_blocks = []
        for block in blocks:
            if not block:
                continue
            data_block = {'header': None, 'loops': [], 'data': {}}
            items = block.split('\n')
            if len(items) == 0:
                continue
            data_block['header'] = StarHeader.from_string('data_' + items[0])
            data, loops = cls._loadBlock('\n'.join(items[1:]))
            for loop in loops:
                data_block['loops'].append(StarLoop.from_data(loop, prefix=prefix))
            for key in data.keys():
                entry = StarEntry.from_string('{}   {}'.format(key, data[key]))
                data_block['data'][entry.name] = entry
            data_blocks.append(data_block)
        if len(data_blocks) == 1:
            data_blocks = data_blocks[0]
        return data_blocks

    @classmethod
    def from_file(cls, filename: str):
        with open(filename, 'r') as reader:
            in_string = reader.read()
        if not in_string:
            in_string = filename
        return cls.from_string(in_string)


class StarSection(StarBase):
    def __str__(self) -> str:
        return self._section_to_string()

    def _section_to_string(self):
        s = ''
        keys = [key for key in self.data[0]._kwargs.keys() if key not in self.exclude]
        for idx, key in enumerate(keys):
            s += f'{StarEntry(self.data[0]._kwargs[key], self.labels[idx], prefix=self.prefix)}\n'
        return s

    def to_class(self, cls, name_conversions=None, skip=[]):
        if not hasattr(cls, 'from_pars'):
            raise AttributeError
        if name_conversions is None:
            name_conversions = [[k1, k2] for k1, k2 in zip(self.labels, self.data[0]._kwargs.keys())]
        new_object = cls.from_pars(**{k[0]: self.data[0]._kwargs[k[1]].raw_value for k in name_conversions})
        for key in name_conversions:
            attr = getattr(new_object, key[0])
            if hasattr(self.data[0]._kwargs[key[1]], 'fixed'):
                if self.data[0]._kwargs[key[1]].fixed is not None:
                    attr.fixed = self.data[0]._kwargs[key[1]].fixed
                if self.data[0]._kwargs[key[1]].error is not None:
                    attr.error = self.data[0]._kwargs[key[1]].error
        return new_object

    @classmethod
    def from_string(cls, in_string: str, name_conversion: List[str] = None, prefix='_'):
        items = in_string.split('\n')
        data = [FakeCore()]
        names = []
        for idx, item in enumerate(items):
            if not item:
                continue
            this_name = None
            if name_conversion is not None:
                this_name = name_conversion[idx]
            conv_item = StarEntry.from_string(item, name_conversion=this_name, prefix=prefix)
            names.append(conv_item.name)
            data[0]._kwargs[conv_item.name] = conv_item.to_fake_item()
        return cls(data, entry_names=names, prefix=prefix)

    @classmethod
    def from_StarEntries(
        cls,
        star_entries: List[StarEntry],
        name_conversions: List[str] = None,
        prefix='_',
    ):
        data = [FakeCore()]
        names = []
        for idx, entry in enumerate(star_entries):
            names.append(entry.name)
            data[0]._kwargs[entry.name] = entry.to_fake_item()
        if name_conversions is None:
            name_conversions = names
        return cls(data, entry_names=name_conversions, prefix=prefix)

    def to_StarEntries(self) -> List[StarEntry]:
        keys = [key for key in self.data[0]._kwargs.keys() if key not in self.exclude]
        return [StarEntry(self.data[0]._kwargs[key], self.labels[idx], prefix=self.prefix) for idx, key in enumerate(keys)]


class StarLoop(StarBase):
    def __str__(self) -> str:
        return self._loop_to_string()

    def _loop_to_string(self):
        s = 'loop_'
        if len(self.data) == 0:
            return ''
        keys = [key for key in self.data[0]._kwargs.keys() if key not in self.exclude]
        for idx, kw in enumerate(keys):
            label = kw
            if not isinstance(self.data[0]._kwargs[kw], FakeItem):
                label = self.data[0]._kwargs[kw].name
            if self.labels[idx] is not None:
                label = self.labels[idx]
            s += '\n {}'.format(self.prefix) + label
        for item in self.data:
            line = '\n'
            for kw in keys:
                val = str(ItemHolder(item._kwargs[kw]))
                if val.count(' ') == 0 and val.count("'") > 0:
                    val = val.strip("'")
                if val[0] == ';':
                    s += line + '\n' + val
                    line = '\n'
                elif len(line) + len(val) + 2 < self.maxlen:
                    line += '  ' + val
                else:
                    s += line
                    line = '\n  ' + val
            s += line
        return s

    @classmethod
    def from_string(cls, in_string: str, name_conversion: List[str] = None, prefix='_'):
        data, loops = cls._loadBlock(in_string, prefix=prefix)
        if len(loops) > 1:
            raise ValueError(f'String has more than one loop: {len(loops)}')
        return cls.from_data(loops[0], name_conversion, prefix)

    @classmethod
    def from_StarSections(
        cls,
        star_sections: List[StarSection],
        name_conversion: List[str] = None,
        prefix='_',
    ):
        this_data = [section.data[0] for section in star_sections]
        all_names = star_sections[0].labels
        if name_conversion is not None:
            all_names = name_conversion
        return cls(this_data, all_names, prefix=prefix)

    def to_StarSections(self) -> List[StarSection]:
        return [StarSection(section, self.labels, prefix=self.prefix) for section in self.data]

    @classmethod
    def from_data(cls, loop: dict, name_conversion: List[str] = None, prefix='_'):
        all_names = []
        all_data = []
        keys = list(loop.keys())
        for idx2 in range(len(loop[keys[0]])):
            fk = FakeCore()
            for idx, key in enumerate(keys):
                this_name = key
                if this_name[0] == '_':
                    this_name = this_name[1:]
                if name_conversion is not None:
                    this_name = name_conversion[idx]
                if idx2 == 0:
                    all_names.append(this_name)
                conv_item = StarEntry.from_string(
                    '{}{}   {}'.format(prefix, this_name, loop[key][idx2]),
                    this_name,
                    prefix=prefix,
                )
                fk._kwargs[conv_item.name] = conv_item.to_fake_item()
            all_data.append(fk)
        return cls(all_data, all_names, prefix=prefix)

    def to_class(self, cls_outer, cls_inner, name_conversions=None):
        if not hasattr(cls_inner, 'from_pars'):
            raise AttributeError
        new_objects = []
        for idx in range(len(self.data)):
            if name_conversions is None:
                keys = [key for key in self.data[idx]._kwargs.keys() if key not in self.exclude]
                name_conversions = [[k, k] for k in keys]
            new_object = cls_inner.from_pars(**{k[0]: self.data[idx]._kwargs[k[1]].raw_value for k in name_conversions})
            for key in name_conversions:
                attr = getattr(new_object, key[0])
                if hasattr(self.data[idx]._kwargs[key[1]], 'fixed'):
                    if self.data[idx]._kwargs[key[1]].fixed is not None:
                        attr.fixed = self.data[idx]._kwargs[key[1]].fixed
                    if self.data[idx]._kwargs[key[1]].error is not None:
                        attr.error = self.data[idx]._kwargs[key[1]].error
            new_objects.append(new_object)
        return cls_outer(cls_outer.__name__, *new_objects)

    def join(self, otherLoop: 'StarLoop', key: str) -> 'StarLoop':
        if key not in self.labels or key not in otherLoop.labels:
            raise AttributeError('Key must be common in both StarLoops')
        if len(self.data) != len(otherLoop.data):
            raise AttributeError('There must be the same number of entries in both StarLoops')
        joint = StarLoop.from_string(str(self))
        for dataset in otherLoop.data:
            lookup_value = dataset._kwargs['label'].raw_value
            try:
                lookup_idx = [d._kwargs['label'].raw_value for d in self.data].index(lookup_value)
            except ValueError:
                raise AttributeError('Both StarLoops must contain the joining same keys')
            joint.data[lookup_idx]._kwargs.update(dataset._kwargs)
        joint.labels.extend([k for k in otherLoop.labels if k != key])
        return joint


class StarHeader:
    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        return 'data_' + self.name

    @classmethod
    def from_string(cls, in_string: str):
        name = in_string.split('data_')[1]
        return cls(name)



# ruff: noqa
__author__ = 'Shyue Ping Ong, Shyam Dwaraknath, Matthew Horton'
__copyright__ = 'Copyright 2011, The Materials Project'
__version__ = '1.0'
__maintainer__ = 'Shyue Ping Ong'
__email__ = 'shyuep@gmail.com'
__status__ = 'Production'
__date__ = 'Sep 23, 2011'

from typing import List
from typing import Tuple
from typing import Union

"""
This module provides classes that operate on points or vectors in 3D space.
"""

import re
import string
import warnings
from math import cos
from math import pi
from math import sin
from math import sqrt

import numpy as np
from easyscience.Objects.core import ComponentSerializer
from easyscience.Utils.string import transformation_to_string


class SymmOp(ComponentSerializer):
    """
    A symmetry operation in cartesian space. Consists of a rotation plus a
    translation. Implementation is as an affine transformation matrix of rank 4
    for efficiency. Read: http://en.wikipedia.org/wiki/Affine_transformation.
    .. attribute:: affine_matrix
        A 4x4 numpy.array representing the symmetry operation.
    """

    _REDIRECT = {'affine_transformation_matrix': lambda obj: getattr(obj, 'affine_matrix').aslist()}

    def __init__(self, affine_transformation_matrix, tol=0.01):
        """
        Initializes the SymmOp from a 4x4 affine transformation matrix.
        In general, this constructor should not be used unless you are
        transferring rotations.  Use the static constructors instead to
        generate a SymmOp from proper rotations and translation.
        Args:
            affine_transformation_matrix (4x4 array): Representing an
                affine transformation.
            tol (float): Tolerance for determining if matrices are equal.
        """
        affine_transformation_matrix = np.array(affine_transformation_matrix)
        if affine_transformation_matrix.shape != (4, 4):
            raise ValueError('Affine Matrix must be a 4x4 numpy array!')
        self.affine_matrix = affine_transformation_matrix
        self.tol = tol

    @staticmethod
    def from_rotation_and_translation(
        rotation_matrix=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
        translation_vec=(0, 0, 0),
        tol=0.1,
    ):
        """
        Creates a symmetry operation from a rotation matrix and a translation
        vector.
        Args:
            rotation_matrix (3x3 array): Rotation matrix.
            translation_vec (3x1 array): Translation vector.
            tol (float): Tolerance to determine if rotation matrix is valid.
        Returns:
            SymmOp object
        """
        rotation_matrix = np.array(rotation_matrix)
        translation_vec = np.array(translation_vec)
        if rotation_matrix.shape != (3, 3):
            raise ValueError('Rotation Matrix must be a 3x3 numpy array.')
        if translation_vec.shape != (3,):
            raise ValueError('Translation vector must be a rank 1 numpy array with 3 elements.')
        affine_matrix = np.eye(4)
        affine_matrix[0:3][:, 0:3] = rotation_matrix
        affine_matrix[0:3][:, 3] = translation_vec
        return SymmOp(affine_matrix, tol)

    def __eq__(self, other):
        return np.allclose(self.affine_matrix, other.affine_matrix, atol=self.tol)

    def __hash__(self):
        return 7

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        output = [
            'Rot:',
            str(self.affine_matrix[0:3][:, 0:3]),
            'tau',
            str(self.affine_matrix[0:3][:, 3]),
        ]
        return '\n'.join(output)

    def operate(self, point):
        """
        Apply the operation on a point.
        Args:
            point: Cartesian coordinate.
        Returns:
            Coordinates of point after operation.
        """
        affine_point = np.array([point[0], point[1], point[2], 1])
        return np.dot(self.affine_matrix, affine_point)[0:3]

    def operate_multi(self, points):
        """
        Apply the operation on a list of points.
        Args:
            points: List of Cartesian coordinates
        Returns:
            Numpy array of coordinates after operation
        """
        points = np.array(points)
        affine_points = np.concatenate([points, np.ones(points.shape[:-1] + (1,))], axis=-1)
        return np.inner(affine_points, self.affine_matrix)[..., :-1]

    def apply_rotation_only(self, vector):
        """
        Vectors should only be operated by the rotation matrix and not the
        translation vector.
        Args:
            vector (3x1 array): A vector.
        """
        return np.dot(self.rotation_matrix, vector)

    def transform_tensor(self, tensor):
        """
        Applies rotation portion to a tensor. Note that tensor has to be in
        full form, not the Voigt form.
        Args:
            tensor (numpy array): a rank n tensor
        Returns:
            Transformed tensor.
        """
        dim = tensor.shape
        rank = len(dim)
        if not all([i == 3 for i in dim]):
            raise ValueError('All dimensions must be equal to 3')

        # Build einstein sum string
        lc = string.ascii_lowercase
        indices = lc[:rank], lc[rank : 2 * rank]
        einsum_string = ','.join([a + i for a, i in zip(*indices)])
        einsum_string += ',{}->{}'.format(*indices[::-1])
        einsum_args = [self.rotation_matrix] * rank + [tensor]

        return np.einsum(einsum_string, *einsum_args)

    def are_symmetrically_related(self, point_a, point_b, tol=0.001):
        """
        Checks if two points are symmetrically related.
        Args:
            point_a (3x1 array): First point.
            point_b (3x1 array): Second point.
            tol (float): Absolute tolerance for checking distance.
        Returns:
            True if self.operate(point_a) == point_b or vice versa.
        """
        if np.allclose(self.operate(point_a), point_b, atol=tol):
            return True
        if np.allclose(self.operate(point_b), point_a, atol=tol):
            return True
        return False

    @property
    def rotation_matrix(self) -> np.ndarray:
        """
        A 3x3 numpy.array representing the rotation matrix.
        """
        return self.affine_matrix[0:3][:, 0:3]

    @property
    def translation_vector(self):
        """
        A rank 1 numpy.array of dim 3 representing the translation vector.
        """
        return self.affine_matrix[0:3][:, 3]

    def __mul__(self, other):
        """
        Returns a new SymmOp which is equivalent to apply the "other" SymmOp
        followed by this one.
        """
        new_matrix = np.dot(self.affine_matrix, other.affine_matrix)
        return SymmOp(new_matrix)

    @property
    def inverse(self):
        """
        Returns inverse of transformation.
        """
        invr = np.linalg.inv(self.affine_matrix)
        return SymmOp(invr)

    @staticmethod
    def from_axis_angle_and_translation(
        axis: Union[Tuple[float, float, float], List[float], np.ndarray],
        angle: float,
        angle_in_radians: bool = False,
        translation_vec: Union[Tuple[float, float, float], List[float], np.ndarray] = (
            0,
            0,
            0,
        ),
    ):
        """
        Generates a SymmOp for a rotation about a given axis plus translation.
        Args:
            axis: The axis of rotation in cartesian space. For example,
                [1, 0, 0] indicates rotation about x-axis.
            angle (float): Angle of rotation.
            angle_in_radians (bool): Set to True if angles are given in
                radians. Or else, units of degrees are assumed.
            translation_vec: A translation vector. Defaults to zero.
        Returns:
            SymmOp for a rotation about given axis and translation.
        """
        if isinstance(axis, (tuple, list)):
            axis = np.array(axis)

        if isinstance(translation_vec, (tuple, list)):
            vec = np.array(translation_vec)
        else:
            vec = translation_vec

        a = angle if angle_in_radians else angle * pi / 180
        cosa = cos(a)
        sina = sin(a)
        u = axis / np.linalg.norm(axis)
        r = np.zeros((3, 3))
        r[0, 0] = cosa + u[0] ** 2 * (1 - cosa)
        r[0, 1] = u[0] * u[1] * (1 - cosa) - u[2] * sina
        r[0, 2] = u[0] * u[2] * (1 - cosa) + u[1] * sina
        r[1, 0] = u[0] * u[1] * (1 - cosa) + u[2] * sina
        r[1, 1] = cosa + u[1] ** 2 * (1 - cosa)
        r[1, 2] = u[1] * u[2] * (1 - cosa) - u[0] * sina
        r[2, 0] = u[0] * u[2] * (1 - cosa) - u[1] * sina
        r[2, 1] = u[1] * u[2] * (1 - cosa) + u[0] * sina
        r[2, 2] = cosa + u[2] ** 2 * (1 - cosa)

        return SymmOp.from_rotation_and_translation(r, vec)

    @staticmethod
    def from_origin_axis_angle(
        origin: Union[Tuple[float, float, float], List[float], np.ndarray],
        axis: Union[Tuple[float, float, float], List[float], np.ndarray],
        angle: float,
        angle_in_radians: bool = False,
    ):
        """
        Generates a SymmOp for a rotation about a given axis through an
        origin.
        Args:
            origin (3x1 array): The origin which the axis passes through.
            axis (3x1 array): The axis of rotation in cartesian space. For
                example, [1, 0, 0]indicates rotation about x-axis.
            angle (float): Angle of rotation.
            angle_in_radians (bool): Set to True if angles are given in
                radians. Or else, units of degrees are assumed.
        Returns:
            SymmOp.
        """
        theta = angle * pi / 180 if not angle_in_radians else angle
        a = origin[0]
        b = origin[1]
        c = origin[2]
        u = axis[0]
        v = axis[1]
        w = axis[2]
        # Set some intermediate values.
        u2 = u * u
        v2 = v * v
        w2 = w * w
        cos_t = cos(theta)
        sin_t = sin(theta)
        l2 = u2 + v2 + w2
        lsq = sqrt(l2)

        # Build the matrix entries element by element.
        m11 = (u2 + (v2 + w2) * cos_t) / l2
        m12 = (u * v * (1 - cos_t) - w * lsq * sin_t) / l2
        m13 = (u * w * (1 - cos_t) + v * lsq * sin_t) / l2
        m14 = (
            a * (v2 + w2) - u * (b * v + c * w) + (u * (b * v + c * w) - a * (v2 + w2)) * cos_t + (b * w - c * v) * lsq * sin_t
        ) / l2

        m21 = (u * v * (1 - cos_t) + w * lsq * sin_t) / l2
        m22 = (v2 + (u2 + w2) * cos_t) / l2
        m23 = (v * w * (1 - cos_t) - u * lsq * sin_t) / l2
        m24 = (
            b * (u2 + w2) - v * (a * u + c * w) + (v * (a * u + c * w) - b * (u2 + w2)) * cos_t + (c * u - a * w) * lsq * sin_t
        ) / l2

        m31 = (u * w * (1 - cos_t) - v * lsq * sin_t) / l2
        m32 = (v * w * (1 - cos_t) + u * lsq * sin_t) / l2
        m33 = (w2 + (u2 + v2) * cos_t) / l2
        m34 = (
            c * (u2 + v2) - w * (a * u + b * v) + (w * (a * u + b * v) - c * (u2 + v2)) * cos_t + (a * v - b * u) * lsq * sin_t
        ) / l2

        return SymmOp(
            [
                [m11, m12, m13, m14],
                [m21, m22, m23, m24],
                [m31, m32, m33, m34],
                [0, 0, 0, 1],
            ]
        )

    @staticmethod
    def reflection(
        normal: Union[Tuple[float, float, float], List[float], np.ndarray],
        origin: Union[Tuple[float, float, float], List[float], np.ndarray] = (0, 0, 0),
    ):
        """
        Returns reflection symmetry operation.
        Args:
            normal (3x1 array): Vector of the normal to the plane of
                reflection.
            origin (3x1 array): A point in which the mirror plane passes
                through.
        Returns:
            SymmOp for the reflection about the plane
        """
        # Normalize the normal vector first.
        n = np.array(normal, dtype=float) / np.linalg.norm(normal)

        u, v, w = n

        translation = np.eye(4)
        translation[0:3, 3] = -np.array(origin)

        xx = 1 - 2 * u**2
        yy = 1 - 2 * v**2
        zz = 1 - 2 * w**2
        xy = -2 * u * v
        xz = -2 * u * w
        yz = -2 * v * w
        mirror_mat = [[xx, xy, xz, 0], [xy, yy, yz, 0], [xz, yz, zz, 0], [0, 0, 0, 1]]

        if np.linalg.norm(origin) > 1e-6:
            mirror_mat = np.dot(np.linalg.inv(translation), np.dot(mirror_mat, translation))
        return SymmOp(mirror_mat)

    @staticmethod
    def inversion(
        origin: Union[Tuple[float, float, float], List[float], np.ndarray] = (0, 0, 0),
    ):
        """
        Inversion symmetry operation about axis.
        Args:
            origin (3x1 array): Origin of the inversion operation. Defaults
                to [0, 0, 0].
        Returns:
            SymmOp representing an inversion operation about the origin.
        """
        mat = -np.eye(4)
        mat[3, 3] = 1
        mat[0:3, 3] = 2 * np.array(origin)
        return SymmOp(mat)

    @staticmethod
    def roto_reflection(
        axis: Union[Tuple[float, float, float], List[float], np.ndarray],
        angle: float,
        origin: Union[Tuple[float, float, float], List[float], np.ndarray] = (0, 0, 0),
    ):
        """
        Returns a roto-reflection symmetry operation
        Args:
            axis (3x1 array): Axis of rotation / mirror normal
            angle (float): Angle in degrees
            origin (3x1 array): Point left invariant by roto-reflection.
                Defaults to (0, 0, 0).
        Return:
            Roto-reflection operation
        """
        rot = SymmOp.from_origin_axis_angle(origin, axis, angle)
        refl = SymmOp.reflection(axis, origin)
        m = np.dot(rot.affine_matrix, refl.affine_matrix)
        return SymmOp(m)

    def as_xyz_string(self) -> str:
        """
        Returns a string of the form 'x, y, z', '-x, -y, z',
        '-y+1/2, x+1/2, z+1/2', etc. Only works for integer rotation matrices
        """
        # test for invalid rotation matrix
        if not np.all(np.isclose(self.rotation_matrix, np.round(self.rotation_matrix))):
            warnings.warn('Rotation matrix should be integer')

        return transformation_to_string(self.rotation_matrix, translation_vec=self.translation_vector, delim=', ')

    @staticmethod
    def from_xyz_string(xyz_string: str):
        """
        Args:
            xyz_string: string of the form 'x, y, z', '-x, -y, z',
                '-2y+1/2, 3x+1/2, z-y+1/2', etc.
        Returns:
            SymmOp
        """
        rot_matrix = np.zeros((3, 3))
        trans = np.zeros(3)
        toks = xyz_string.strip().replace(' ', '').lower().split(',')
        re_rot = re.compile(r'([+-]?)([\d\.]*)/?([\d\.]*)([x-z])')
        re_trans = re.compile(r'([+-]?)([\d\.]+)/?([\d\.]*)(?![x-z])')
        for i, tok in enumerate(toks):
            # build the rotation matrix
            for m in re_rot.finditer(tok):
                factor = -1 if m.group(1) == '-' else 1
                if m.group(2) != '':
                    factor *= float(m.group(2)) / float(m.group(3)) if m.group(3) != '' else float(m.group(2))
                j = ord(m.group(4)) - 120
                rot_matrix[i, j] = factor
            # build the translation vector
            for m in re_trans.finditer(tok):
                factor = -1 if m.group(1) == '-' else 1
                num = float(m.group(2)) / float(m.group(3)) if m.group(3) != '' else float(m.group(2))
                trans[i] = num * factor
        return SymmOp.from_rotation_and_translation(rot_matrix, trans)



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from .atoms import Atoms
from .lattice import Lattice
from .spacegroup import SpaceGroup

__all__ = ['Atoms', 'Lattice', 'SpaceGroup']



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from __future__ import annotations

import collections
import itertools
import math
import warnings
from copy import deepcopy
from fractions import Fraction
from functools import reduce
from typing import TYPE_CHECKING
from typing import ClassVar
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import numpy as np
from easyscience import ureg
from easyscience.Constraints import ObjConstraint
from easyscience.Objects.ObjectClasses import BaseObj
from easyscience.Objects.ObjectClasses import Parameter
from easyscience.Utils.decorators import memoized

from .SpaceGroup import SpaceGroup

Vector3Like = Union[List[float], np.ndarray]


if TYPE_CHECKING:
    from easyscience.Utils.typing import iF

CELL_DETAILS = {
    'length': {
        'description': 'Unit-cell length of the selected structure in angstroms.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_cell/',
        'value': 3,
        'units': 'angstrom',
        'min': 0,
        'max': np.Inf,
        'fixed': True,
    },
    'angle': {
        'description': 'Unit-cell angle of the selected structure in degrees.',
        'url': 'https://docs.easydiffraction.org/lib/dictionaries/_cell/',
        'value': 90,
        'units': 'deg',
        'min': 0,
        'max': np.Inf,
        'fixed': True,
    },
}


class Lattice(BaseObj):
    _REDIRECT = {'ang_unit': None}

    length_a: ClassVar[Parameter]
    length_b: ClassVar[Parameter]
    length_c: ClassVar[Parameter]
    angle_alpha: ClassVar[Parameter]
    angle_beta: ClassVar[Parameter]
    angle_gamma: ClassVar[Parameter]

    def __init__(
        self,
        length_a: Optional[Union[Parameter, float]] = None,
        length_b: Optional[Union[Parameter, float]] = None,
        length_c: Optional[Union[Parameter, float]] = None,
        angle_alpha: Optional[Union[Parameter, float]] = None,
        angle_beta: Optional[Union[Parameter, float]] = None,
        angle_gamma: Optional[Union[Parameter, float]] = None,
        interface: Optional[iF] = None,
        ang_unit: str = 'deg',
        **kwargs,
    ):
        THESE_CELL_DETAILS = deepcopy(CELL_DETAILS)
        THESE_CELL_DETAILS['angle']['units'] = ang_unit

        super().__init__(
            'lattice',
            length_a=Parameter('length_a', **THESE_CELL_DETAILS['length']),
            length_b=Parameter('length_b', **THESE_CELL_DETAILS['length']),
            length_c=Parameter('length_c', **THESE_CELL_DETAILS['length']),
            angle_alpha=Parameter('angle_alpha', **THESE_CELL_DETAILS['angle']),
            angle_beta=Parameter('angle_beta', **THESE_CELL_DETAILS['angle']),
            angle_gamma=Parameter('angle_gamma', **THESE_CELL_DETAILS['angle']),
            **kwargs,
        )
        if length_a is not None:
            self.length_a = length_a
        if length_b is not None:
            self.length_b = length_b
        if length_c is not None:
            self.length_c = length_c

        if angle_alpha is not None:
            self.angle_alpha = angle_alpha
        if angle_beta is not None:
            self.angle_beta = angle_beta
        if angle_gamma is not None:
            self.angle_gamma = angle_gamma

        if THESE_CELL_DETAILS['angle']['units'] is not CELL_DETAILS['angle']['units']:
            self.angle_alpha.convert_unit(CELL_DETAILS['angle']['units'])
            self.angle_beta.convert_unit(CELL_DETAILS['angle']['units'])
            self.angle_gamma.convert_unit(CELL_DETAILS['angle']['units'])

        self.interface = interface

    @classmethod
    def from_matrix(
        cls,
        matrix: Union[List[float], List[List[float]], np.ndarray],
        interface: Optional[iF] = None,
    ) -> L:
        """
        Construct a crystallographic unit cell from the lattice matrix

        :param matrix: Matrix describing the crystallographic unit cell in the form of a numpy array,
        1x9 list or 3x3 list.
        :type matrix: np.ndarray, List[float], List[List[float]]
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """

        matrix = np.array(matrix, dtype=np.float64).reshape((3, 3))
        lengths = np.sqrt(np.sum(matrix**2, axis=1))
        angles = np.zeros(3)
        for i in range(3):
            j = (i + 1) % 3
            k = (i + 2) % 3
            angles[i] = np.dot(matrix[j], matrix[k]) / (lengths[j] * lengths[k])
            angles[i] = max(min(angles[i], 1), -1)
        angles = np.arccos(angles) * 180.0 / np.pi
        return cls(*lengths, *angles, interface=interface)

    @classmethod
    def cubic(cls, a: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a cubic lattice.

        :param a: The *a* lattice parameter of the cubic cell.
        :type a: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, a, 90, 90, 90, interface=interface)

    @classmethod
    def tetragonal(cls, a: float, c: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a tetragonal lattice.

        :param a: *a* lattice parameter of the tetragonal cell.
        :type a: float
        :param c: *c* lattice parameter of the tetragonal cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, c, 90, 90, 90, interface=interface)

    @classmethod
    def orthorhombic(cls, a: float, b: float, c: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for an orthorhombic lattice.

        :param a: *a* lattice parameter of the orthorhombic cell.
        :type a: float
        :param b: *b* lattice parameter of the orthorhombic cell.
        :type b: float
        :param c: *c* lattice parameter of the orthorhombic cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, b, c, 90, 90, 90, interface=interface)

    @classmethod
    def monoclinic(
        cls,
        a: float,
        b: float,
        c: float,
        beta: float,
        interface: Optional[iF] = None,
    ) -> L:
        """
        Convenience constructor for a monoclinic lattice of dimensions a x b x c with non right-angle
        beta between lattice vectors a and c.

        :param a: *a* lattice parameter of the monoclinc cell.
        :type a: float
        :param b: *b* lattice parameter of the monoclinc cell.
        :type b: float
        :param c: *c* lattice parameter of the monoclinc cell.
        :type c: float
        :param beta: *beta* angle between lattice vectors b and c in degrees.
        :type beta: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, b, c, 90, beta, 90, interface=interface)

    @classmethod
    def hexagonal(cls, a: float, c: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a hexagonal lattice of dimensions a x a x c

        :param a: *a* lattice parameter of the hexagonal cell.
        :type a: float
        :param c: *c* lattice parameter of the hexagonal cell.
        :type c: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, c, 90, 90, 120, interface=interface)

    @classmethod
    def rhombohedral(cls, a: float, alpha: float, interface: Optional[iF] = None) -> L:
        """
        Convenience constructor for a rhombohedral lattice of dimensions a x a x a.

        :param a: *a* lattice parameter of the monoclinc cell.
        :type a: float
        :param alpha: Angle for the rhombohedral lattice in degrees.
        :type alpha: float
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        return cls(a, a, a, alpha, alpha, alpha, interface=interface)

    # Dynamic properties
    @property
    def a(self) -> float:
        """
        Get the *a* lattice parameter.

        :return: *a* lattice parameter
        :rtype: float
        """
        return self.length_a.raw_value

    @a.setter
    def a(self, new_a_value: float):
        """
        Set the *a* lattice parameter.

        :param new_a_value: new *a* lattice parameter
        :type new_a_value: float
        :return: noneType
        :rtype: None
        """
        self.length_a.value = new_a_value

    @property
    def b(self) -> float:
        """
        Get the *b* lattice parameter.

        :return: *b* lattice parameter
        :rtype: float
        """
        return self.length_b.raw_value

    @b.setter
    def b(self, new_b_value: float):
        """
        Set the *b* lattice parameter.

        :param new_b_value: new *a* lattice parameter
        :type new_b_value: float
        :return: noneType
        :rtype: None
        """
        self.length_b.value = new_b_value

    @property
    def c(self) -> float:
        """
        Get the *c* lattice parameter.

        :return: *c* lattice parameter
        :rtype: float
        """
        return self.length_c.raw_value

    @c.setter
    def c(self, new_c_value: float):
        """
        Set the *c* lattice parameter.

        :param new_c_value: new *a* lattice parameter
        :type new_c_value: float
        :return: noneType
        :rtype: None
        """
        self.length_c.value = new_c_value

    @property
    def alpha(self) -> float:
        """
        Get the *alpha* lattice parameter.

        :return: *alpha* lattice parameter
        :rtype: float
        """
        return self.angle_alpha.raw_value

    @alpha.setter
    def alpha(self, new_alpha_value: float):
        """
        Set the *alpha* lattice parameter.

        :param new_alpha_value: new *alpha* lattice parameter
        :type new_alpha_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_alpha.value = new_alpha_value

    @property
    def beta(self) -> float:
        """
        Get the *beta* lattice parameter.

        :return: *beta* lattice parameter
        :rtype: float
        """
        return self.angle_beta.raw_value

    @beta.setter
    def beta(self, new_beta_value: float):
        """
        Set the *beta* lattice parameter.

        :param new_beta_value: new *a* lattice parameter
        :type new_beta_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_beta.value = new_beta_value

    @property
    def gamma(self) -> float:
        """
        Get the *gamma* lattice parameter.

        :return: *gamma* lattice parameter
        :rtype: float
        """
        return self.angle_gamma.raw_value

    @gamma.setter
    def gamma(self, new_gamma_value: float):
        """
        Set the *gamma* lattice parameter.

        :param new_gamma_value: new *gamma* lattice parameter
        :type new_gamma_value: float
        :return: noneType
        :rtype: None
        """
        self.angle_gamma.value = new_gamma_value

    @property
    def lengths(self) -> Tuple[float, float, float]:
        """
        Get the lengths of the unit cell.

        :return: Tuple of unit cell lengths (a, b, c)
        :rtype: tuple
        """
        return self.a, self.b, self.c

    @property
    def angles(self) -> Tuple[float, float, float]:
        """
        Get the angles of the unit cell in degrees.

        :return: Tuple of unit cell angles (alpha, beta, gamma)
        :rtype: tuple
        """
        return self.alpha, self.beta, self.gamma

    @property
    def matrix(self) -> np.ndarray:
        """
        Get the lattice matrix.

        :return: Lattice matrix in the form of a 9x9 matrix
        :rtype: np.ndarray
        """
        return self.__matrix(*self.lengths, *self.angles)

    @property
    def volume(self) -> float:
        """
        Volume of the unit cell.

        :return: Volume of the unit cell
        :rtype: ureg.Quantity
        """
        m = self.matrix
        vol = float(abs(np.dot(np.cross(m[0], m[1]), m[2])))
        return ureg.Quantity(vol, units=self.length_a.unit * self.length_b.unit * self.length_c.unit)

    @property
    def inv_matrix(self) -> np.ndarray:
        """
        Inverse of lattice matrix.

        :return: Inverse of lattice matrix
        :rtype: np.ndarray
        """
        return np.linalg.inv(self.matrix)

    @property
    def metric_tensor(self) -> np.ndarray:
        """
        The metric tensor of the lattice.

        :return metric tensor of the lattice
        :rtype: np.ndarray
        """
        matrix = self.matrix
        return np.dot(matrix, self.matrix.T)

    # Functions that create new copies
    @property
    def reciprocal_lattice(self) -> L:
        """
        Return the reciprocal lattice. Note that this is the standard
        reciprocal lattice used for solid state physics with a factor of 2 *
        pi. If you are looking for the crystallographic reciprocal lattice,
        use the reciprocal_lattice_crystallographic property.

        :return: New cell in the reciprocal lattice
        :rtype: Lattice
        """
        v = np.linalg.inv(self.matrix).T
        return self.__class__.from_matrix(v * 2 * np.pi, interface=self.interface)

    @property
    def reciprocal_lattice_crystallographic(self) -> L:
        """
        Returns the *crystallographic* reciprocal lattice, i.e., no factor of
        2 * pi.

        :return: New cell in the *crystallographic* reciprocal lattice
        :rtype: Lattice
        """
        return self.__class__.from_matrix(self.reciprocal_lattice.matrix / (2 * np.pi), interface=self.interface)

    def scale(self, new_volume: float) -> L:
        """
        Return a new Cell with volume new_volume by performing a
        scaling of the lattice vectors so that length proportions and angles
        are preserved.

        :param new_volume: New volume to scale to.
        :type new_volume: float
        :return: New cell scaled to volume
        :rtype: Lattice
        """

        lengths = self.lengths
        versors = self.matrix / lengths
        geo_factor = np.abs(np.dot(np.cross(versors[0], versors[1]), versors[2]))
        ratios = np.array(lengths) / lengths[2]
        new_c = (new_volume / (geo_factor * np.prod(ratios))) ** (1 / 3.0)
        return self.__class__.from_matrix(versors * (new_c * ratios), interface=self.interface)

    def scale_lengths(self, length_scales: Union[float, Vector3Like]) -> L:
        """
        Return a new Cell where the cell lengths have been scaled by the scaling
        factors. Angles are preserved.

        :param length_scales: Scaling for each length or universal scaling factor
        :type length_scales: Union[float, Vector3Like]
        :return: New Cell scaled by user supplied scale factors
        :rtype: Lattice
        """
        if isinstance(length_scales, float):
            length_scales = 3 * [length_scales]
        new_lengths = np.array(length_scales) * np.array(self.lengths)
        return self.__class__(*new_lengths, *self.angles, interface=self.interface)

    # Get functions
    def get_cartesian_coords(self, fractional_coords: Vector3Like) -> np.ndarray:
        """
        Returns the cartesian coordinates given fractional coordinates.

        :param fractional_coords: fractional_coords (3x1 array): Fractional coords.
        :type fractional_coords: np.ndarray , list
        :return: Cartesian coordinates
        :rtype: np.ndarray
        """
        return np.dot(fractional_coords, self.matrix)

    def get_fractional_coords(self, cart_coords: Vector3Like) -> np.ndarray:
        """
        Returns the fractional coordinates given cartesian coordinates.

        :param cart_coords: cart_coords (3x1 array): Fractional coords.
        :type cart_coords: np.ndarray , list
        :return: Fractional coordinates.
        :rtype: np.ndarray
        """
        return np.dot(cart_coords, self.inv_matrix)

    def get_vector_along_lattice_directions(self, cart_coords: Vector3Like) -> np.ndarray:
        """
        Returns the coordinates along lattice directions given cartesian coordinates.
        Note, this is different than a projection of the cartesian vector along the
        lattice parameters. It is simply the fractional coordinates multiplied by the
        lattice vector magnitudes.

        :param cart_coords: cart_coords (3x1 array): Fractional coords.
        :type cart_coords: np.ndarray , list
        :return: Lattice coordinates.
        :rtype: np.ndarray
        """
        return self.lengths * self.get_fractional_coords(cart_coords)

    def d_hkl(self, miller_index: Vector3Like) -> float:
        """
        Returns the distance between the hkl plane and the origin

        :param miller_index: (h. k. l) Miller index of plane
        :type miller_index: Vector3Like
        :return: Distance between the hkl plane and the origin
        :rtype: float
        """

        gstar = self.reciprocal_lattice_crystallographic.metric_tensor
        hkl = np.array(miller_index)
        return 1 / ((np.dot(np.dot(hkl, gstar), hkl.T)) ** (1 / 2))

    # Checking
    def is_orthogonal(self) -> bool:
        """
        Returns true if the cell is orthogonal (all angles 90 degrees)

        :return: Whether all angles are 90 degrees.
        :rtype: bool
        """
        return all([abs(a - 90) < 1e-5 for a in self.angles])

    def is_hexagonal(self, hex_angle_tol: float = 5, hex_length_tol: float = 0.01) -> bool:
        """
        Returns true if the Lattice is hexagonal.

        :param hex_angle_tol: Angle tolerance
        :param hex_length_tol: Length tolerance
        :return: Whether lattice corresponds to hexagonal lattice.
        :rtype: bool
        """
        lengths = self.lengths
        angles = self.angles
        right_angles = [i for i in range(3) if abs(angles[i] - 90) < hex_angle_tol]
        hex_angles = [i for i in range(3) if abs(angles[i] - 60) < hex_angle_tol or abs(angles[i] - 120) < hex_angle_tol]

        return (
            len(right_angles) == 2
            and len(hex_angles) == 1
            and abs(lengths[right_angles[0]] - lengths[right_angles[1]]) < hex_length_tol
        )

    @staticmethod
    @memoized
    def __matrix(a: float, b: float, c: float, alpha: float, beta: float, gamma: float) -> np.ndarray:
        """
        Calculating the crystallographic matrix is time consuming and we use it often, so we have memoized it.
        :param a: *a* lattice parameter
        :type a: float
        :param b: *b* lattice parameter
        :type b: float
        :param c: *c* lattice parameter
        :type c: float
        :param alpha: *alpha* lattice parameter
        :type alpha: float
        :param beta: *beta* lattice parameter
        :type beta: float
        :param gamma: *gamma* lattice parameter
        :type gamma: float
        :return: crystallographic matrix
        :rtype: np.ndarray
        """
        angles_r = np.radians([alpha, beta, gamma])
        cos_alpha, cos_beta, cos_gamma = np.cos(angles_r)
        sin_alpha, sin_beta, sin_gamma = np.sin(angles_r)

        val = (cos_alpha * cos_beta - cos_gamma) / (sin_alpha * sin_beta)
        # Sometimes rounding errors result in values slightly > 1.
        val = max(min(val, 1), -1)
        gamma_star = np.arccos(val)

        vector_a = [a * sin_beta, 0.0, a * cos_beta]
        vector_b = [
            -b * sin_alpha * np.cos(gamma_star),
            b * sin_alpha * np.sin(gamma_star),
            b * cos_alpha,
        ]
        vector_c = [0.0, 0.0, float(c)]
        return np.array([vector_a, vector_b, vector_c], dtype=np.float64)

    # noinspection PyStringFormat
    def __repr__(self) -> str:
        return (
            '<Lattice: (a: {:.2f} {:~P}, b: {:.2f} {:~P}, c: {:.2f}{:~P}, alpha: {:.2f} {:~P}, beta: {:.2f} {:~P}, '
            ''
            'gamma: {:.2f} {:~P}>'.format(
                self.length_a.raw_value,
                self.length_a.unit,
                self.length_b.raw_value,
                self.length_b.unit,
                self.length_c.raw_value,
                self.length_c.unit,
                self.angle_alpha.raw_value,
                self.angle_alpha.unit,
                self.angle_beta.raw_value,
                self.angle_beta.unit,
                self.angle_gamma.raw_value,
                self.angle_gamma.unit,
            )
        )

    def __format__(self, fmt_spec=''):
        """
        Support format printing. Supported formats are:

        1. "l" for a list format that can be easily copied and pasted, e.g.,
           ".3fl" prints something like
           "[[10.000, 0.000, 0.000], [0.000, 10.000, 0.000], [0.000, 0.000, 10.000]]"
        2. "p" for lattice parameters ".1fp" prints something like
           "{10.0, 10.0, 10.0, 90.0, 90.0, 90.0}"
        3. Default will simply print a 3x3 matrix form. E.g.,
           10.000 0.000 0.000
           0.000 10.000 0.000
           0.000 0.000 10.000
        """
        m = (self.lengths, self.angles)

        if fmt_spec.endswith('m'):
            fmt = '[[{}, {}, {}], [{}, {}, {}], [{}, {}, {}]]'
            m = self.matrix.tolist()
            fmt_spec = fmt_spec[:-1]
        elif fmt_spec.endswith('l'):
            fmt = '{{{}, {}, {}, {}, {}, {}}}'
            fmt_spec = fmt_spec[:-1]
        else:
            fmt = '({} {} {}), ({} {} {})'
            fmt_spec = fmt_spec[:-1]
        return fmt.format(*[format(c, fmt_spec) for row in m for c in row])

    def __copy__(self):
        """
        Returns a deep copy of the Lattice


        :return: Deep copy of self
        :rtype:
        """
        return self.__class__(*self.lengths, *self.angles, interface=self.interface)

    def get_points_in_sphere(
        self,
        frac_points: List[Vector3Like],
        center: Vector3Like,
        r: float,
        zip_results=True,
    ) -> Union[
        List[Tuple[np.ndarray, float, int, np.ndarray]],
        List[np.ndarray],
    ]:
        """
        Find all points within a sphere from the point taking into account
        periodic boundary conditions. This includes sites in other periodic
        images.
        Algorithm:
        1. place sphere of radius r in crystal and determine minimum supercell
           (parallelpiped) which would contain a sphere of radius r. for this
           we need the projection of a_1 on a unit vector perpendicular
           to a_2 & a_3 (i.e. the unit vector in the direction b_1) to
           determine how many a_1"s it will take to contain the sphere.
           Nxmax = r * length_of_b_1 / (2 Pi)
        2. keep points falling within r.
        Args:
            frac_points: All points in the lattice in fractional coordinates.
            center: Cartesian coordinates of center of sphere.
            r: radius of sphere.
            zip_results (bool): Whether to zip the results together to group by
                 point, or return the raw fcoord, dist, index arrays
        Returns:
            if zip_results:
                [(fcoord, dist, index, supercell_image) ...] since most of the time, subsequent
                processing requires the distance, index number of the atom, or index of the image
            else:
                fcoords, dists, inds, image
        """
        cart_coords = self.get_cartesian_coords(frac_points)
        neighbors = get_points_in_spheres(
            all_coords=cart_coords,
            center_coords=np.array([center]),
            r=r,
            pbc=True,
            numerical_tol=1e-8,
            lattice=self,
            return_fcoords=True,
        )[0]
        if len(neighbors) < 1:
            return [] if zip_results else [()] * 4
        if zip_results:
            return neighbors
        return [np.array(i) for i in list(zip(*neighbors))]


L = TypeVar('L', bound=Lattice)


class PeriodicLattice(Lattice):
    spacegroup: ClassVar[SpaceGroup]

    def __init__(
        self,
        length_a: Optional[Union[Parameter, float]] = None,
        length_b: Optional[Union[Parameter, float]] = None,
        length_c: Optional[Union[Parameter, float]] = None,
        angle_alpha: Optional[Union[Parameter, float]] = None,
        angle_beta: Optional[Union[Parameter, float]] = None,
        angle_gamma: Optional[Union[Parameter, float]] = None,
        spacegroup: Optional[Union[SpaceGroup, str]] = None,
        interface: Optional[iF] = None,
    ):
        super().__init__(
            length_a=length_a,
            length_b=length_b,
            length_c=length_c,
            angle_alpha=angle_alpha,
            angle_beta=angle_beta,
            angle_gamma=angle_gamma,
            spacegroup=SpaceGroup('P1'),
        )
        if spacegroup is not None:
            if isinstance(spacegroup, str):
                spacegroup = SpaceGroup(spacegroup)
            self.spacegroup = spacegroup

        # Do some class voodoo. We can do this as we have hidden space_group_HM_name
        # as _space_group_HM_name and used a simple property.
        self.__previous_SG_setter = spacegroup.__class__.space_group_HM_name.fset
        spacegroup.__class__.space_group_HM_name = property(
            fget=spacegroup.__class__.space_group_HM_name.fget,
            fset=lambda obj, val: self.__new_SG_setter(obj, val),
            fdel=spacegroup.__class__.space_group_HM_name.fdel,
        )

        self.interface = interface
        self.enforce_sym()

    @classmethod
    def from_lattice_and_spacegroup(cls: Type[L], lattice: Lattice, spacegroup: SpaceGroup):
        length_a = lattice.length_a.raw_value
        length_b = lattice.length_b.raw_value
        length_c = lattice.length_c.raw_value
        angle_alpha = lattice.angle_alpha.raw_value
        angle_beta = lattice.angle_beta.raw_value
        angle_gamma = lattice.angle_gamma.raw_value

        return cls(
            length_a,
            length_b,
            length_c,
            angle_alpha,
            angle_beta,
            angle_gamma,
            spacegroup,
            interface=lattice.interface,
        )

    @classmethod
    def from_matrix(
        cls: Type[L],
        matrix: Union[List[float], List[List[float]], np.ndarray],
        interface: Optional[iF] = None,
    ) -> L:
        """
        Construct a crystallographic unit cell from the lattice matrix

        :param matrix: Matrix describing the crystallographic unit cell in the form of a numpy array,
        1x9 list or 3x3 list.
        :type matrix: np.ndarray, List[float], List[List[float]]
        :return: Crystallographic unit cell container
        :rtype: Lattice
        """
        raise NotImplementedError('A periodic Lattice can not be created from just a matrix')

    def clear_sym(self):
        pars = [
            self.length_a,
            self.length_b,
            self.length_c,
            self.angle_alpha,
            self.angle_beta,
            self.angle_gamma,
        ]
        for par in pars:
            new_con = {con: par.user_constraints[con] for con in par.user_constraints.keys() if not con.startswith('sg_')}
            par.user_constraints = new_con
            if not par.enabled:
                par.enabled = True

    def enforce_sym(self):
        """
        Enforce symmetry constraints on to a Lattice
        """
        # SG system
        crys_system = self.spacegroup.crystal_system
        self.clear_sym()
        trig_test = crys_system == 'trigonal' and (
            self.spacegroup.setting_str.endswith('H')
            or self.spacegroup.int_number
            in [
                143,
                144,
                145,
                147,
                149,
                150,
                151,
                152,
                153,
                154,
                156,
                157,
                158,
                159,
                162,
                163,
                164,
                165,
            ]
        )

        # Go through the cell systems
        if crys_system == 'cubic':
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.length_a.user_constraints['sg_2'] = ObjConstraint(self.length_c, '', self.length_a)
            self.length_a.user_constraints['sg_2']()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == 'hexagonal' or trig_test:
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 120
            self.angle_gamma.enabled = False
        elif crys_system == 'trigonal' and not trig_test:
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.length_a.user_constraints['sg_2'] = ObjConstraint(self.length_c, '', self.length_a)
            self.length_a.user_constraints['sg_2']()
            self.angle_alpha.user_constraints['sg_1'] = ObjConstraint(self.angle_beta, '', self.angle_alpha)
            self.angle_alpha.user_constraints['sg_1']()
            self.angle_alpha.user_constraints['sg_2'] = ObjConstraint(self.angle_gamma, '', self.angle_alpha)
            self.angle_alpha.user_constraints['sg_2']()
        elif crys_system == 'tetragonal':
            self.length_a.user_constraints['sg_1'] = ObjConstraint(self.length_b, '', self.length_a)
            self.length_a.user_constraints['sg_1']()
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == 'orthorhombic':
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_beta = 90
            self.angle_beta.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == 'monoclinic':
            self.angle_alpha = 90
            self.angle_alpha.enabled = False
            self.angle_gamma = 90
            self.angle_gamma.enabled = False
        elif crys_system == 'triclinic':
            return
        else:
            raise TypeError('The current crystal system is unknown so symmetry cannot be enforced')

    def to_cell(self) -> L:
        """
        Convert the PeriodicLattice to a standard Lattice. i.e. drop spacegroup. The return is a copy

        :return: Lattice with spacegroup information dropped.
        :rtype: Lattice
        """
        return Lattice(*self.lengths, *self.angles, interface=None)

    def __copy__(self):
        """
        Returns a deep copy of the Periodic-Lattice. Note that the spacegroup parameter is also a copy!!

        :return: Deep copy of self
        :rtype: PeriodicLattice
        """
        return self.__class__(
            *self.lengths,
            *self.angles,
            self.spacegroup.hermann_mauguin,
            interface=self.interface,
        )

    def __new_SG_setter(self, obj, value):
        self.clear_sym()
        self.__previous_SG_setter(obj, value)
        self.enforce_sym()


def get_integer_index(miller_index: Sequence[float], round_dp: int = 4, verbose: bool = True) -> Tuple[int, int, int]:
    """
    Attempt to convert a vector of floats to whole numbers.
    Args:
        miller_index (list of float): A list miller indexes.
        round_dp (int, optional): The number of decimal places to round the
            miller index to.
        verbose (bool, optional): Whether to print warnings.
    Returns:
        (tuple): The Miller index.
    """
    mi = np.asarray(miller_index)
    # deal with the case we have small irregular floats
    # that are all equal or factors of each other
    mi /= min([m for m in mi if m != 0])
    mi /= np.max(np.abs(mi))

    # deal with the case we have nice fractions
    md = [Fraction(n).limit_denominator(12).denominator for n in mi]
    mi *= reduce(lambda x, y: x * y, md)
    int_miller_index = np.int_(np.round(mi, 1))
    mi /= np.abs(reduce(math.gcd, int_miller_index))

    # round to a reasonable precision
    mi = np.array([round(h, round_dp) for h in mi])

    # need to recalculate this after rounding as values may have changed
    int_miller_index = np.int_(np.round(mi, 1))
    if np.any(np.abs(mi - int_miller_index) > 1e-6) and verbose:
        warnings.warn('Non-integer encountered in Miller index')
    else:
        mi = int_miller_index

    # minimise the number of negative indexes
    mi += 0  # converts -0 to 0

    def n_minus(index):
        return len([h for h in index if h < 0])

    if n_minus(mi) > n_minus(mi * -1):
        mi *= -1

    # if only one index is negative, make sure it is the smallest
    # e.g. (-2 1 0) -> (2 -1 0)
    if sum(mi != 0) == 2 and n_minus(mi) == 1 and abs(min(mi)) > max(mi):
        mi *= -1

    return tuple(mi)  # type: ignore


def get_points_in_spheres(
    all_coords: np.ndarray,
    center_coords: np.ndarray,
    r: float,
    pbc: Union[bool, List[bool]] = True,
    numerical_tol: float = 1e-8,
    lattice: Lattice = None,
    return_fcoords: bool = False,
) -> List[List[Tuple[np.ndarray, float, int, np.ndarray]]]:
    """
    For each point in `center_coords`, get all the neighboring points in `all_coords` that are within the
    cutoff radius `r`.
    Args:
        all_coords: (list of cartesian coordinates) all available points
        center_coords: (list of cartesian coordinates) all centering points
        r: (float) cutoff radius
        pbc: (bool or a list of bool) whether to set periodic boundaries
        numerical_tol: (float) numerical tolerance
        lattice: (Lattice) lattice to consider when PBC is enabled
        return_fcoords: (bool) whether to return fractional coords when pbc is set.
    Returns:
        List[List[Tuple[coords, distance, index, image]]]
    """
    if isinstance(pbc, bool):
        pbc = [pbc] * 3
    pbc = np.array(pbc, dtype=bool)
    if return_fcoords and lattice is None:
        raise ValueError('Lattice needs to be supplied to compute fractional coordinates')
    center_coords_min = np.min(center_coords, axis=0)
    center_coords_max = np.max(center_coords, axis=0)
    # The lower bound of all considered atom coords
    global_min = center_coords_min - r - numerical_tol
    global_max = center_coords_max + r + numerical_tol
    if np.any(pbc):
        if lattice is None:
            raise ValueError('Lattice needs to be supplied when considering periodic boundary')
        recp_len = np.array(lattice.reciprocal_lattice.lengths)
        maxr = np.ceil((r + 0.15) * recp_len / (2 * math.pi))
        frac_coords = lattice.get_fractional_coords(center_coords)
        nmin_temp = np.floor(np.min(frac_coords, axis=0)) - maxr
        nmax_temp = np.ceil(np.max(frac_coords, axis=0)) + maxr
        nmin = np.zeros_like(nmin_temp)
        nmin[pbc] = nmin_temp[pbc]
        nmax = np.ones_like(nmax_temp)
        nmax[pbc] = nmax_temp[pbc]
        all_ranges = [np.arange(x, y, dtype='int64') for x, y in zip(nmin, nmax)]
        matrix = lattice.matrix
        # temporarily hold the fractional coordinates
        image_offsets = lattice.get_fractional_coords(all_coords)
        all_fcoords = []
        # only wrap periodic boundary
        for k in range(3):
            if pbc[k]:  # type: ignore
                all_fcoords.append(np.mod(image_offsets[:, k : k + 1], 1))
            else:
                all_fcoords.append(image_offsets[:, k : k + 1])
        all_fcoords = np.concatenate(all_fcoords, axis=1)
        image_offsets = image_offsets - all_fcoords
        coords_in_cell = np.dot(all_fcoords, matrix)
        # Filter out those beyond max range
        valid_coords = []
        valid_images = []
        valid_indices = []
        for image in itertools.product(*all_ranges):
            coords = np.dot(image, matrix) + coords_in_cell
            valid_index_bool = np.all(
                np.bitwise_and(coords > global_min[None, :], coords < global_max[None, :]),
                axis=1,
            )
            ind = np.arange(len(all_coords))
            if np.any(valid_index_bool):
                valid_coords.append(coords[valid_index_bool])
                valid_images.append(np.tile(image, [np.sum(valid_index_bool), 1]) - image_offsets[valid_index_bool])
                valid_indices.extend([k for k in ind if valid_index_bool[k]])
        if len(valid_coords) < 1:
            return [[]] * len(center_coords)
        valid_coords = np.concatenate(valid_coords, axis=0)
        valid_images = np.concatenate(valid_images, axis=0)

    else:
        valid_coords = all_coords
        valid_images = [[0, 0, 0]] * len(valid_coords)
        valid_indices = np.arange(len(valid_coords))

    # Divide the valid 3D space into cubes and compute the cube ids
    all_cube_index = _compute_cube_index(valid_coords, global_min, r)
    nx, ny, nz = _compute_cube_index(global_max, global_min, r) + 1
    all_cube_index = _three_to_one(all_cube_index, ny, nz)
    site_cube_index = _three_to_one(_compute_cube_index(center_coords, global_min, r), ny, nz)
    # create cube index to coordinates, images, and indices map
    cube_to_coords = collections.defaultdict(list)  # type: Dict[int, List]
    cube_to_images = collections.defaultdict(list)  # type: Dict[int, List]
    cube_to_indices = collections.defaultdict(list)  # type: Dict[int, List]
    for i, j, k, m in zip(all_cube_index.ravel(), valid_coords, valid_images, valid_indices):
        cube_to_coords[i].append(j)
        cube_to_images[i].append(k)
        cube_to_indices[i].append(m)

    # find all neighboring cubes for each atom in the lattice cell
    site_neighbors = find_neighbors(site_cube_index, nx, ny, nz)
    neighbors = []  # type: List[List[Tuple[np.ndarray, float, int, np.ndarray]]]

    for i, j in zip(center_coords, site_neighbors):
        l1 = np.array(_three_to_one(j, ny, nz), dtype=int).ravel()
        # use the cube index map to find the all the neighboring
        # coords, images, and indices
        ks = [k for k in l1 if k in cube_to_coords]
        if not ks:
            neighbors.append([])
            continue
        nn_coords = np.concatenate([cube_to_coords[k] for k in ks], axis=0)
        nn_images = itertools.chain(*[cube_to_images[k] for k in ks])
        nn_indices = itertools.chain(*[cube_to_indices[k] for k in ks])
        dist = np.linalg.norm(nn_coords - i[None, :], axis=1)
        nns: List[Tuple[np.ndarray, float, int, np.ndarray]] = []
        for coord, index, image, d in zip(nn_coords, nn_indices, nn_images, dist):
            # filtering out all sites that are beyond the cutoff
            # Here there is no filtering of overlapping sites
            if d < r + numerical_tol:
                if return_fcoords and (lattice is not None):
                    coord = np.round(lattice.get_fractional_coords(coord), 10)
                nn = (coord, float(d), int(index), image)
                nns.append(nn)
        neighbors.append(nns)
    return neighbors


# The following internal methods are used in the get_points_in_sphere method.
def _compute_cube_index(coords: np.ndarray, global_min: float, radius: float) -> np.ndarray:
    """
    Compute the cube index from coordinates
    Args:
        coords: (nx3 array) atom coordinates
        global_min: (float) lower boundary of coordinates
        radius: (float) cutoff radius
    Returns: (nx3 array) int indices
    """
    return np.array(np.floor((coords - global_min) / radius), dtype=int)


def _one_to_three(label1d: np.ndarray, ny: int, nz: int) -> np.ndarray:
    """
    Convert a 1D index array to 3D index array
    Args:
        label1d: (array) 1D index array
        ny: (int) number of cells in y direction
        nz: (int) number of cells in z direction
    Returns: (nx3) int array of index
    """
    last = np.mod(label1d, nz)
    second = np.mod((label1d - last) / nz, ny)
    first = (label1d - last - second * nz) / (ny * nz)
    return np.concatenate([first, second, last], axis=1)


def _three_to_one(label3d: np.ndarray, ny: int, nz: int) -> np.ndarray:
    """
    The reverse of _one_to_three
    """
    return np.array(label3d[:, 0] * ny * nz + label3d[:, 1] * nz + label3d[:, 2]).reshape((-1, 1))


def find_neighbors(label: np.ndarray, nx: int, ny: int, nz: int) -> List[np.ndarray]:
    """
    Given a cube index, find the neighbor cube indices
    Args:
        label: (array) (n,) or (n x 3) indice array
        nx: (int) number of cells in y direction
        ny: (int) number of cells in y direction
        nz: (int) number of cells in z direction
    Returns: neighbor cell indices
    """

    array = [[-1, 0, 1]] * 3
    neighbor_vectors = np.array(list(itertools.product(*array)), dtype=int)
    if np.shape(label)[1] == 1:
        label3d = _one_to_three(label, ny, nz)
    else:
        label3d = label
    all_labels = label3d[:, None, :] - neighbor_vectors[None, :, :]
    filtered_labels = []
    # filter out out-of-bound labels i.e., label < 0
    for labels in all_labels:
        ind = (labels[:, 0] < nx) * (labels[:, 1] < ny) * (labels[:, 2] < nz) * np.all(labels > -1e-5, axis=1)
        filtered_labels.append(labels[ind])
    return filtered_labels



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

import pytest

from easyscience import global_object
from easyscience.models.polynomial import Line
from easyscience.Objects.Groups import BaseCollection
from easyscience.Objects.Variable import Descriptor
from easyscience.Objects.Variable import Parameter
from easycrystallography.Components.Site import Atoms, Site, _SITE_DETAILS
from easycrystallography.io.star_base import ItemHolder
from easycrystallography.io.star_base import StarLoop
from easycrystallography.io.star_base import StarSection
import gc


@pytest.mark.parametrize(
    "value, error, precision, expected",
    (
        (1.234560e05, 1.230000e02, 1, "123500(100)"),
        (1.234567e01, 1.230000e-03, 2, "12.3457(12)"),
        (1.234560e-01, 1.230000e-04, 3, "0.123456(123)"),
        (1.234560e-03, 1.234500e-08, 4, "0.00123456000(1234)"),
        (1.234560e-05, 1.234000e-07, 1, "0.0000123(1)"),
    ),
    ids=[
        "1.234560e+05 +- 1.230000e+02 @1",
        "1.234567e+01 +- 1.230000e-03 @2",
        "1.234560e-01 +- 1.230000e-04 @3",
        "1.234560e-03 +- 1.234500e-08 @4",
        "1.234560e-05 +- 1.234000e-07 @1",
    ],
)
def test_ItemHolder_with_error(value, error, precision, expected):
    p = Parameter("p", value, error=error)
    s = ItemHolder(p, decimal_places=precision)
    assert str(s) == expected


@pytest.mark.parametrize("fixed", (True, False), ids=["fixed", "not fixed"])
@pytest.mark.parametrize(
    "value, precision, expected",
    (
        (1.234560e05, 1, "123456.0"),
        (1.234567e01, 2, "12.35"),
        (1.234560e-01, 3, "0.123"),
        (1.234560e-03, 4, "0.0012"),
        (1.234560e-05, 1, "0.0"),
        (1.234560e-05, 5, "0.00001"),
    ),
    ids=[
        "1.234560e+05 @1",
        "1.234567e+01 @2",
        "1.234560e-01 @3",
        "1.234560e-03 @4",
        "1.234560e-05 @1",
        "1.234560e-05 @5",
    ],
)
# FAILED tests/unit_tests/io/test_star.py::test_ItemHolder_fixed[1.234560e-01 @3-fixed] - RuntimeError: dictionary changed size during iteration
# Adding map clear because of these errors happening ONLY in 3.12
def test_ItemHolder_fixed(fixed, value, precision, expected):
    global_object.map._clear()
    p = Parameter("p", value, fixed=fixed)
    s = ItemHolder(p, decimal_places=precision)
    if not p.fixed:
        expected += "()"
    assert str(s) == expected


@pytest.mark.parametrize("cls", [Descriptor])
def test_ItemHolder_str(cls):
    v = cls("v", "fooooooooo")
    s = ItemHolder(v)
    assert str(s) == "fooooooooo"


def test_StarSection():
    l = Line(2, 3)
    s = StarSection(l)
    expected = "_m   2.00000000()\n_c   3.00000000()\n"
    assert str(s) == expected


def test_StarLoop():
    gc.collect()
    l1 = Line(2, 3)
    l2 = Line(4, 5)

    ps = BaseCollection("LineCollection", l1, l2)
    s = StarLoop(ps)

    expected = (
        "loop_\n _m\n _c\n  2.00000000()  3.00000000()\n  4.00000000()  5.00000000()"
    )

    assert str(s) == expected



# SPDX-FileCopyrightText: 2024 EasyCrystallography contributors
# SPDX-License-Identifier: BSD-3-Clause
# © 2022-2024 Contributors to the EasyCrystallography project <https://github.com/EasyScience/EasyCrystallography>

from typing import List

import pytest
import numpy as np
from easycrystallography.Components.Site import Atoms, Site, _SITE_DETAILS

site_details = [Site('Al', 'Al'), Site('Fe', 'Fe3+'), Site('TEST', 'H')]


def gen_sites() -> List[Site]:
    args = []
    for edx in range(len(site_details)):
        key = f'from_{edx+1}_sites'
        args.append(pytest.param((site_details[0:(edx + 1)]), id=key))
    return args


def test_Atoms_empty():
    name = 'test'
    atoms = Atoms(name)
    assert len(atoms) == 0
    assert atoms.name == name


def test_Atoms_wrong_name():
    name = False
    with pytest.raises(TypeError):
        atoms = Atoms(name)


# def test_Atoms_wrong_args():
#     name = 'test'
#     args = [False, 'testing', (), []]
#     with pytest.raises(TypeError):
#         atoms = Atoms(name, *args)

@pytest.mark.parametrize('sites', gen_sites())
def test_Atoms_creation_args(sites: List[Site]):

    name = 'test'
    atoms = Atoms(name, *sites)
    assert len(atoms) == len(sites)
    assert atoms.name == name


@pytest.mark.parametrize('sites', gen_sites())
def test_Atoms_creation_dict(sites: List[Site]):

    name = 'test'
    d = {}
    for idx, site in enumerate(sites):
        d[str(idx)] = site
    atoms = Atoms(name, **d)
    assert len(atoms) == len(sites)
    assert atoms.name == name


@pytest.mark.parametrize('sites', gen_sites())
def test_Atoms_repr(sites: List[Site]):
    name = 'test'
    atoms = Atoms(name, *sites)
    assert str(atoms) == f'Collection of {len(atoms)} sites.'


@pytest.mark.parametrize('sites', gen_sites())
def test_Atoms_get_item_int(sites: List[Site]):
    name = 'test'
    atoms = Atoms(name, *sites)

    for atom, site in zip(atoms, sites):
        assert atom.label == site.label

    for idx, site in enumerate(sites):
        atom = atoms[idx]
        assert atom.label == site.label


@pytest.mark.parametrize('sites', gen_sites())
def test_Atoms_get_item_str(sites: List[Site]):
    name = 'test'
    atoms = Atoms(name, *sites)

    for site in sites:
        atom = atoms[site.label.raw_value]
        assert atom.label == site.label